"""
Dataclass Wizard
~~~~~~~~~~~~~~~~

Lightning-fast JSON wizardry for Python dataclasses — effortless
serialization right out of the box!

Sample Usage:

    >>> from dataclasses import dataclass, field
    >>> from datetime import datetime
    >>> from typing import Optional
    >>>
    >>> from dataclass_wizard import JSONWizard
    >>> from dataclass_wizard.properties import property_wizard
    >>>
    >>>
    >>> @dataclass
    >>> class MyClass(JSONWizard, metaclass=property_wizard):
    >>>
    >>>     my_str: Optional[str]
    >>>     list_of_int: list[int] = field(default_factory=list)
    >>>     # You can also define this as `my_dt`, however only the annotation
    >>>     # will carry over in that case, since the value is re-declared by
    >>>     # the property below.
    >>>     _my_dt: datetime = datetime(2000, 1, 1)
    >>>
    >>>     @property
    >>>     def my_dt(self):
    >>>     # A sample `getter` which returns the datetime with year set as 2010
    >>>         if self._my_dt is not None:
    >>>             return self._my_dt.replace(year=2010)
    >>>         return self._my_dt
    >>>
    >>>     @my_dt.setter
    >>>     def my_dt(self, new_dt: datetime):
    >>>     # A sample `setter` which sets the inverse (roughly) of
    >>>     # the `month` and `day`
    >>>         self._my_dt = new_dt.replace(month=13 - new_dt.month,
    >>>                                      day=30 - new_dt.day)
    >>>
    >>>
    >>> string = '''{"myStr": 42, "listOFInt": [1, "2", 3]}'''
    >>> c = MyClass.from_json(string)
    >>> print(repr(c))
    >>> # prints:
    >>> #   MyClass(
    >>> #       my_str='42',
    >>> #       list_of_int=[1, 2, 3],
    >>> #       my_dt=datetime.datetime(2010, 12, 29, 0, 0)
    >>> #   )
    >>> my_dict = {'My_Str': 'string', 'myDT': '2021-01-20T15:55:30Z'}
    >>> c = MyClass.from_dict(my_dict)
    >>> print(repr(c))
    >>> # prints:
    >>> #   MyClass(
    >>> #       my_str='string',
    >>> #       list_of_int=[],
    >>> #       my_dt=datetime.datetime(2010, 12, 10, 15, 55, 30,
    >>> #                               tzinfo=datetime.timezone.utc)
    >>> #   )
    >>> print(c.to_json())
    >>> # prints:
    >>> #   {"myStr": "string", "listOfInt": [], "myDt": "2010-12-10T15:55:30Z"}

For full documentation and more advanced usage, please see
<https://dcw.ritviknag.com>.

:copyright: (c) 2021-2026 by Ritvik Nag.
:license: Apache 2.0, see LICENSE for more details.
"""
from logging import NullHandler

from ._log import LOG
from ._public import *

# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
LOG.addHandler(NullHandler())



"""
Dict helper module

TODO: Delete when time allows --
  See https://github.com/rnag/dataclass-wizard/issues/215
"""


class NestedDict(dict):
    """
    A dictionary that automatically creates nested dictionaries for missing keys.

    This class extends the built-in `dict` to simplify working with deeply nested structures.
    If a key is accessed but does not exist, it will be created automatically with a new `NestedDict` as its value.

    Source: https://stackoverflow.com/a/5369984/10237506

    Example:
        >>> nd = NestedDict()
        >>> nd['a']['b']['c'] = 42
        >>> nd
        {'a': {'b': {'c': 42}}}

        >>> nd['x']['y']
        {}
    """

    __slots__ = ()

    def __getitem__(self, key):
        """
        Retrieve the value for a key, or create a nested dictionary for missing keys.

        Args:
            key (Hashable): The key to retrieve or create.

        Returns:
            Any: The value associated with the key, or a new `NestedDict` for missing keys.

        Example:
            >>> nd = NestedDict()
            >>> nd['foo']  # Creates a new NestedDict for 'foo'
            {}

        Note:
            If the key exists, its value is returned. Otherwise, a new `NestedDict` is created,
            stored, and returned.
        """
        if key in self: return self.get(key)
        return self.setdefault(key, NestedDict())



import pytest

from typing import Literal

from dataclass_wizard.v0.parsers import LiteralParser


class TestLiteralParser:
    @pytest.fixture
    def literal_parser(self) -> LiteralParser:
        return LiteralParser(cls=object, base_type=Literal["foo"], extras={})

    def test_literal_parser_dunder_contains_succeeds_if_item_in_keys_of_base_type(self, literal_parser):
        assert "foo" in literal_parser

    def test_literal_parser_dunder_contains_fails_if_item_not_in_keys_of_base_type(self, literal_parser):
        assert "bar" not in literal_parser



__all__ = [
    'Buffer',
    'Unpack',
    'PyForwardRef',
    'PyProtocol',
    'PyDeque',
    'PyTypedDict',
    'PyRequired',
    'PyNotRequired',
    'PyReadOnly',
    'PyLiteralString',
    'FrozenKeys',
    'DefFactory',
    'NoneType',
    'ExplicitNullType',
    'ExplicitNull',
    'JSONList',
    'JSONObject',
    'ListOfJSONObject',
    'JSONValue',
    'FileType',
    'EnvFileType',
    'StrCollection',
    'ParseFloat',
    'Encoder',
    'FileEncoder',
    'Decoder',
    'FileDecoder',
    'NUMBERS',
    'T',
    'E',
    'U',
    'M',
    'NT',
    'DT',
    'DD',
    'N',
    'S',
    'LT',
    'LSQ',
    'FREF',
    'dataclass_transform',
]

from collections import defaultdict, deque
from collections.abc import Collection, Iterable, Mapping, Sequence
from datetime import date, datetime, time
from enum import Enum
from os import PathLike
from typing import (
    Any,
    AnyStr,
    BinaryIO,
    Callable,
    NamedTuple,
    TextIO,
    TypeVar,
    Union,
)
from typing import (
    Deque as PyDeque,
)
from typing import (
    ForwardRef as PyForwardRef,
)
from typing import (
    Protocol as PyProtocol,
)
from typing import (
    TypedDict as PyTypedDict,
)
from uuid import UUID

from .constants import (
    PY310_OR_ABOVE,
    PY311_OR_ABOVE,
    PY312_OR_ABOVE,
    PY313_OR_ABOVE,
)

# The class of the `None` singleton, cached for re-usability
if PY310_OR_ABOVE:
    # https://docs.python.org/3/library/types.html#types.NoneType
    from types import NoneType
else:
    # "Cannot assign to a type"
    NoneType = type(None)  # type: ignore[misc]

# Type check for numeric types - needed because `bool` is technically
# a Number.
NUMBERS = int, float

# Generic type
T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)
TT = TypeVar('TT')

# Enum subclass type
E = TypeVar('E', bound=Enum)

# UUID subclass type
U = TypeVar('U', bound=UUID)

# Mapping type
M = TypeVar('M', bound=Mapping)

# NamedTuple type
NT = TypeVar('NT', bound=NamedTuple)

# Date, time, or datetime type
DT = TypeVar('DT', date, time, datetime)

# DefaultDict type
DD = TypeVar('DD', bound=defaultdict)

# Numeric type
N = Union[int, float]

# Sequence type
S = TypeVar('S', bound=Sequence)

# List or Tuple type
LT = TypeVar('LT', list, tuple)

# List, Set, or Deque (Double ended queue) type
LSQ = TypeVar('LSQ', list, set, frozenset, deque)

# A fixed set of key names
FrozenKeys = frozenset[str]

# Default factory type, assuming a no-args constructor
DefFactory = Callable[[], T]

# Valid collection types in JSON.
JSONList = list[Any]
JSONObject = dict[str, Any]
ListOfJSONObject = list[JSONObject]

# Valid value types in JSON.
JSONValue = Union[None, str, bool, int, float, JSONList, JSONObject]

# File-type argument, compatible with the type of `file` for `open`
FileType = Union[str, bytes, PathLike, int]

# DotEnv file-type argument (string, tuple of string, boolean, or None)
EnvFileType = Union[bool, FileType, Iterable[FileType], None]

# Type for a string or a collection of strings.
StrCollection = Union[str, Collection[str]]

# Python 3.11 introduced `Required` and `NotRequired` wrappers for
# `TypedDict` fields (PEP 655). Python 3.9+ users can import the
# wrappers from `typing_extensions`.

if PY313_OR_ABOVE:  # pragma: no cover
    from collections.abc import Buffer
    from typing import LiteralString as PyLiteralString
    from typing import NotRequired as PyNotRequired
    from typing import ReadOnly as PyReadOnly
    from typing import Required as PyRequired
    from typing import Unpack, dataclass_transform
elif PY311_OR_ABOVE:  # pragma: no cover
    if PY312_OR_ABOVE:
        from collections.abc import Buffer
    else:
        from typing_extensions import Buffer

    from typing import LiteralString as PyLiteralString
    from typing import NotRequired as PyNotRequired
    from typing import Required as PyRequired
    from typing import Unpack, dataclass_transform

    from typing_extensions import ReadOnly as PyReadOnly
else:
    from typing_extensions import Buffer, Unpack, dataclass_transform
    from typing_extensions import LiteralString as PyLiteralString
    from typing_extensions import NotRequired as PyNotRequired
    from typing_extensions import ReadOnly as PyReadOnly
    from typing_extensions import Required as PyRequired

# Forward references can be either strings or explicit `ForwardRef` objects.
# noinspection SpellCheckingInspection
FREF = TypeVar('FREF', str, PyForwardRef)


class _UnsetType:
    __slots__ = ()
    def __repr__(self) -> str:
        return 'UNSET'

UNSET = _UnsetType()


# runtime placeholders so "from x import META" works
META = ENV_META = type


class ExplicitNullType:
    __slots__ = ()  # Saves memory by preventing the creation of instance dictionaries

    # Class-level instance variable for singleton control
    _instance: "ExplicitNullType | None" = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __bool__(self):
        return False

    def __repr__(self):
        return '<ExplicitNull>'


# Create the singleton instance
ExplicitNull = ExplicitNullType()

# Type annotations
ParseFloat = Callable[[str], Any]


class Encoder(PyProtocol):
    """
    Represents an encoder for Python object -> JSON, e.g. analogous to
    `json.dumps`
    """

    def __call__(self, obj,
                 /,
                 *args,
                 **kwargs) -> str:
        ...


class FileEncoder(PyProtocol):
    """
    Represents an encoder for Python object -> JSON file, e.g. analogous to
    `json.dump`
    """

    def __call__(self, obj, file,
                 /,
                 *args,
                 **kwargs) -> None:
        ...


class Decoder(PyProtocol):
    """
    Represents a decoder for JSON -> Python object, e.g. analogous to
    `json.loads`
    """

    def __call__(self, s: AnyStr,
                 **kwargs) -> Union[JSONObject, ListOfJSONObject]:
        ...


class FileDecoder(PyProtocol):
    """
    Represents a decoder for JSON file -> Python object, e.g. analogous to
    `json.load`
    """
    def __call__(self, file: Union[TextIO, BinaryIO],
                 **kwargs) -> Union[JSONObject, ListOfJSONObject]:
        ...



import json
import logging
from dataclasses import dataclass, MISSING

from .abstractions import AbstractJSONWizard
from .bases_meta import BaseJSONWizardMeta, LoadMeta, DumpMeta, register_type
from .class_helper import call_meta_initializer_if_needed
from .constants import PACKAGE_NAME
from .loader_selection import asdict, fromdict, fromlist
from .log import enable_library_debug_logging
from .type_def import dataclass_transform
# noinspection PyProtectedMember
from .utils.dataclass_compat import (_create_fn,
                                     _dataclass_needs_refresh,
                                     _set_new_attribute)


def _str_fn():
    return _create_fn('__str__',
                      ('self',),
                      ['return self.to_json(indent=2)'])


def _first_declared_attr_in_mro(cls, name: str):
    """First `name` found in MRO (excluding cls); else None."""
    for base in cls.__mro__[1:]:
        attr = base.__dict__.get(name, MISSING)
        if attr is not MISSING:
            return attr
    return None


def _set_from_dict_and_to_dict_if_needed(cls):
    """
    Pin default dispatchers on subclasses.

    Codegen is lazy; if a base later gets a specialised
    `from_dict` / `to_dict`, subclasses would inherit it.
    Defining defaults in `cls.__dict__` blocks that.
    """
    if 'from_dict' not in cls.__dict__:
        inherited = _first_declared_attr_in_mro(cls, 'from_dict')
        if getattr(inherited, '__func__', None) is fromdict:
            cls.from_dict = classmethod(fromdict)

    if 'to_dict' not in cls.__dict__:
        inherited = _first_declared_attr_in_mro(cls, 'to_dict')
        if inherited is asdict:
            cls.to_dict = asdict


# noinspection PyShadowingBuiltins
def _configure_wizard_class(cls,
                            str=True,
                            debug=False,
                            case=None,
                            dump_case=None,
                            load_case=None,
                            _key_transform=None):
    load_meta_kwargs = {}

    # if case is not None:
    #     _v1_default = True
    #     load_meta_kwargs['v1_case'] = case
    #
    # if dump_case is not None:
    #     load_meta_kwargs['v1_dump_case'] = dump_case
    #
    # if load_case is not None:
    #     load_meta_kwargs['v1_load_case'] = load_case

    if _key_transform is not None:
        DumpMeta(key_transform=_key_transform).bind_to(cls)

    if debug:
        # minimum logging level for logs by this library
        lvl = logging.DEBUG if isinstance(debug, bool) else debug
        # enable library logging
        enable_library_debug_logging(lvl)
        # set `debug_enabled` flag for the class's Meta
        load_meta_kwargs['debug_enabled'] = lvl

    if load_meta_kwargs:
        LoadMeta(**load_meta_kwargs).bind_to(cls)

    # Calls the Meta initializer when inner :class:`Meta` is sub-classed.
    call_meta_initializer_if_needed(cls)

    # Add a `__str__` method to the subclass, if needed
    if str:
        _set_new_attribute(cls, '__str__', _str_fn())

    # Add `from_dict` and `to_dict` methods to the subclass, if needed
    _set_from_dict_and_to_dict_if_needed(cls)


@dataclass_transform()
class DataclassWizard(AbstractJSONWizard):

    __slots__ = ()

    class Meta(BaseJSONWizardMeta):

        __slots__ = ()

        __is_inner_meta__ = True

        def __init_subclass__(cls):
            return cls._init_subclass()

    register_type = classmethod(register_type)

    @classmethod
    def from_json(cls, string, *,
                  decoder=json.loads,
                  **decoder_kwargs):

        o = decoder(string, **decoder_kwargs)

        return fromdict(cls, o) if isinstance(o, dict) else fromlist(cls, o)

    from_list = classmethod(fromlist)

    from_dict = classmethod(fromdict)

    to_dict = asdict

    def to_json(self, *,
                encoder=json.dumps,
                **encoder_kwargs):

        return encoder(asdict(self), **encoder_kwargs)

    @classmethod
    def list_to_json(cls,
                     instances,
                     encoder=json.dumps,
                     **encoder_kwargs):

        list_of_dict = [asdict(o, cls=cls) for o in instances]

        return encoder(list_of_dict, **encoder_kwargs)

    # noinspection PyShadowingBuiltins
    def __init_subclass__(cls,
                          str=False,
                          debug=False,
                          case=None,
                          dump_case=None,
                          load_case=None,
                          _key_transform=None,
                          _apply_dataclass=True,
                          **dc_kwargs):

        super().__init_subclass__()

        # skip classes provided by this library.
        if cls.__module__.startswith(f'{PACKAGE_NAME}.'):
            return

        # Apply the @dataclass decorator.
        if _apply_dataclass and _dataclass_needs_refresh(cls):
            # noinspection PyArgumentList
            dataclass(cls, **dc_kwargs)

        _configure_wizard_class(cls, str, debug, case, dump_case, load_case,
                                _key_transform)


# noinspection PyAbstractClass
class JSONSerializable(DataclassWizard):

    __slots__ = ()

    # noinspection PyShadowingBuiltins
    def __init_subclass__(cls,
                          str=True,
                          debug=False,
                          case=None,
                          dump_case=None,
                          load_case=None,
                          _key_transform=None,
                          _apply_dataclass=False,
                          **_):

        super().__init_subclass__(str, debug, case, dump_case, load_case,
                                  _key_transform, _apply_dataclass)


def _str_pprint_fn():
    from pprint import pformat

    def __str__(self):
        return pformat(self, width=70)

    return __str__


# A handy alias in case it comes in useful to anyone :)
JSONWizard = JSONSerializable


class JSONPyWizard(JSONWizard):
    """Helper for JSONWizard that ensures dumping to JSON keeps keys as-is."""

    # noinspection PyShadowingBuiltins
    def __init_subclass__(cls,
                          str=True,
                          debug=False,
                          case=None,
                          dump_case=None,
                          load_case=None,
                          _key_transform=None,
                          _apply_dataclass=False,
                          **_):
        """Bind child class to DumpMeta with no key transformation."""

        # Call JSONSerializable.__init_subclass__()
        # set `key_transform_with_dump` for the class's Meta
        super().__init_subclass__(False, debug, case, dump_case, load_case, 'NONE',
                                  _apply_dataclass)

        # Add a `__str__` method to the subclass, if needed
        if str:
            _set_new_attribute(cls, '__str__', _str_pprint_fn())



"""
Lazy Import definitions. Generally, these imports will be available when any
"bonus features" are installed, i.e. as below:

  $ pip install dataclass-wizard[timedelta]
"""

from .constants import PY311_OR_ABOVE
from .utils._lazy_loader import LazyLoader

# python-dotenv: for loading environment values from `.env` files
dotenv = LazyLoader(globals(), 'dotenv', 'dotenv', local_name='python-dotenv')

# pytimeparse: for parsing JSON string values as a `datetime.timedelta`
pytimeparse = LazyLoader(globals(), 'pytimeparse', 'timedelta')

# PyYAML: to add support for (de)serializing YAML data to dataclass instances
yaml = LazyLoader(globals(), 'yaml', 'yaml', local_name='PyYAML')

# Tomli -or- tomllib (PY 3.11+): to add support for (de)serializing TOML
# data to dataclass instances
if PY311_OR_ABOVE:
    import tomllib as toml
else:
    toml = LazyLoader(globals(), 'tomli', 'toml', local_name='tomli')

# Tomli-W: to add support for serializing dataclass instances to TOML
toml_w = LazyLoader(globals(), 'tomli_w', 'toml', local_name='tomli-w')



import enum
import typing

from _typeshed import Incomplete

from .utils._string_case import to_camel_case as to_camel_case
from .utils._string_case import to_lisp_case as to_lisp_case
from .utils._string_case import to_pascal_case as to_pascal_case
from .utils._string_case import to_snake_case as to_snake_case

class FuncWrapper:
    f: Incomplete
    def __init__(self, f: typing.Callable) -> None: ...
    def __call__(self, *args, **kwargs): ...

class KeyAction(enum.Enum):
    IGNORE = ...
    RAISE = ...
    WARN = ...
    @staticmethod
    def _generate_next_value_(name, start, count, last_values): ...
    @classmethod
    def __init__(cls, value) -> None: ...

class EnvKeyStrategy(enum.Enum):
    ENV = ...
    FIELD_FIRST = ...
    STRICT = ...
    @staticmethod
    def _generate_next_value_(name, start, count, last_values): ...
    @classmethod
    def __init__(cls, value) -> None: ...

class KeyCase(enum.Enum):
    CAMEL = ...
    C = ...
    PASCAL = ...
    P = ...
    KEBAB = ...
    K = ...
    SNAKE = ...
    S = ...
    AUTO = ...
    A = ...
    @staticmethod
    def _generate_next_value_(name, start, count, last_values): ...
    def __call__(self, *args): ...
    @classmethod
    def __init__(cls, value) -> None: ...

class DateTimeTo(enum.Enum):
    ISO = ...
    TIMESTAMP = ...
    @staticmethod
    def _generate_next_value_(name, start, count, last_values): ...
    @classmethod
    def __init__(cls, value) -> None: ...

class EnvPrecedence(enum.Enum):
    SECRETS_ENV_DOTENV = ...
    SECRETS_DOTENV_ENV = ...
    ENV_ONLY = ...
    @staticmethod
    def _generate_next_value_(name, start, count, last_values): ...
    @classmethod
    def __init__(cls, value) -> None: ...




def per_cls(cache, cls, factory=dict):
    # returns the per-class dict, creating if absent
    value = cache.get(cls)
    if value is None:
        value = cache[cls] = factory()
    return value


def is_builtin(o):

    # Fast path: check if object is a builtin singleton
    # TODO replace with `match` statement once we drop support for Python 3.9
    # match x:
    #     case None: pass
    #     case True: pass
    #     case False: pass
    #     case builtins.Ellipsis: pass
    if o in {None, True, False, ...}:
        return True

    return getattr(o, '__class__', o).__module__ == 'builtins'


def create_new_class(
        class_or_instance, bases,
        suffix=None, attr_dict=None):

    if not suffix and bases:
        suffix = get_class_name(bases[0])

    new_cls_name = f'{get_class_name(class_or_instance)}{suffix}'

    return type(
        new_cls_name,
        bases,
        attr_dict or {'__slots__': ()}
    )


def get_class_name(class_or_instance):

    try:
        return class_or_instance.__qualname__
    except AttributeError:
        # We're dealing with a dataclass instance
        return type(class_or_instance).__qualname__


def get_outer_class_name(inner_cls, default=None, raise_=True):

    try:
        name = get_class_name(inner_cls).rsplit('.', 1)[-2]
        # This is mainly for our test cases, where we nest the class
        # definition in the test func. Either way, it's not a valid class.
        assert not name.endswith('<locals>')

    except (IndexError, AssertionError):
        if raise_:
            raise
        return default

    else:
        return name


def get_class(obj):

    return obj if isinstance(obj, type) else type(obj)


def is_subclass(obj, base_cls):

    cls = obj if isinstance(obj, type) else type(obj)
    return issubclass(cls, base_cls)


def is_subclass_safe(cls, class_or_tuple):

    try:
        return issubclass(cls, class_or_tuple)
    except TypeError:
        return False



"""
Ideally should be in the `bases` module, however we'll run into a Circular
Import scenario if we move it there, since the `loaders` and `dumpers` modules
both import directly from `bases`.

"""
from collections.abc import Mapping, Sequence
from datetime import tzinfo
from typing import Any, Callable, Literal, TypeAlias, TypeVar

from ._bases import AbstractEnvMeta, AbstractMeta, TypeToHook
from ._loaders import LoadMixin
from ._models import Extras, TypeInfo
from ._path_util import EnvFilePaths, SecretsDirs
from ._type_def import ENV_META, META, E
from .conditions import Condition
from .constants import TAG
from .enums import DateTimeTo, EnvKeyStrategy, EnvPrecedence, KeyAction, KeyCase

ALLOWED_MODES = Literal['runtime', 'codegen']

# global flag to determine if debug mode was ever enabled
_debug_was_enabled = False

HookFn = Callable[..., Any]

L = TypeVar('L', bound=LoadMixin)

# (cls, container_tp, tp, extras) -> new_tp
PreDecoder: TypeAlias = Callable[[L, type | None, TypeInfo, Extras], TypeInfo]


def register_type(cls, tp: type, *,
                  load: HookFn | None = None,
                  dump: HookFn | None = None,
                  mode: str | None = None) -> None: ...


def _enable_debug_mode_if_needed(cls_loader, possible_lvl: bool | int | str):
    ...


def _as_enum_safe(cls: type, name: str, base_type: type[E]) -> E | None:
    ...


class BaseJSONWizardMeta(AbstractMeta):

    __slots__ = ()

    @classmethod
    def _init_subclass(cls):
        ...

    @classmethod
    def bind_to(cls, dataclass: type, create=True, is_default=True,
                base_loader=None, base_dumper=None):
        ...


class BaseEnvWizardMeta(AbstractEnvMeta):

    __slots__ = ()

    @classmethod
    def _init_subclass(cls):
        ...

    @classmethod
    def bind_to(cls, env_class: type, create=True, is_default=True):
        ...


# noinspection PyPep8Naming
def LoadMeta(*,
             debug: bool | int | str = ...,
             recursive: bool = True,
             tag: str = ...,
             tag_key: str = TAG,
             auto_assign_tags: bool = ...,
             type_to_hook: TypeToHook = ...,
             pre_decoder: PreDecoder = ...,
             case: KeyCase | str | None = ...,
             field_to_alias: Mapping[str, str | Sequence[str]] = ...,
             on_unknown_key: KeyAction | str | None = KeyAction.IGNORE,
             unsafe_parse_dataclass_in_union: bool = ...,
             namedtuple_as_dict: bool = ...,
             coerce_none_to_empty_str: bool = ...,
             leaf_handling: Literal['exact', 'issubclass'] = ...) -> META:
    ...


# noinspection PyPep8Naming
def DumpMeta(*,
             debug: bool | int | str = ...,
             recursive: bool = True,
             tag: str = ...,
             skip_defaults: bool = ...,
             skip_if: Condition = ...,
             skip_defaults_if: Condition = ...,
             type_to_hook: TypeToHook = ...,
             case: KeyCase | str | None = ...,
             field_to_alias: Mapping[str, str | Sequence[str]] = ...,
             dump_date_time_as: DateTimeTo | str = ...,
             assume_naive_datetime_tz: tzinfo | None = ...,
             namedtuple_as_dict: bool = ...,
             leaf_handling: Literal['exact', 'issubclass'] = ...) -> META:
    ...


# noinspection PyPep8Naming
def EnvMeta(*,
            debug: bool | int | str = ...,
            recursive: bool = True,
            env_file: EnvFilePaths = ...,
            env_prefix: str = ...,
            secrets_dir: SecretsDirs = ...,
            skip_defaults: bool = ...,
            skip_if: Condition = ...,
            skip_defaults_if: Condition = ...,
            tag: str = ...,
            tag_key: str = TAG,
            auto_assign_tags: bool = ...,
            type_to_load_hook: TypeToHook = ...,
            type_to_dump_hook: TypeToHook = ...,
            pre_decoder: PreDecoder = ...,
            load_case: EnvKeyStrategy | str = ...,
            dump_case: KeyCase | str = ...,
            env_precedence: EnvPrecedence = ...,
            field_to_env_load: Mapping[str, str | Sequence[str]] = ...,
            field_to_alias_dump: Mapping[str, str | Sequence[str]] = ...,
            # on_unknown_key: KeyAction | str | None = KeyAction.IGNORE,
            unsafe_parse_dataclass_in_union: bool = ...,
            dump_date_time_as: DateTimeTo | str = ...,
            assume_naive_datetime_tz: tzinfo | None = ...,
            namedtuple_as_dict: bool = ...,
            coerce_none_to_empty_str: bool = ...,
            leaf_handling: Literal['exact', 'issubclass'] = ...) -> ENV_META:
    ...



import typing
from datetime import tzinfo
from typing import Callable
from typing import ClassVar as _ClassVar

from ._bases_meta import ALLOWED_MODES, HookFn, PreDecoder
from ._decorators import cached_class_property as cached_class_property
from ._path_util import EnvFilePaths, SecretsDirs
from ._type_def import META
from .conditions import Condition
from .enums import DateTimeTo as DateTimeTo
from .enums import EnvKeyStrategy as EnvKeyStrategy
from .enums import EnvPrecedence as EnvPrecedence
from .enums import KeyAction as KeyAction
from .enums import KeyCase as KeyCase

TypeToHook = typing.Mapping[
    type, tuple[ALLOWED_MODES, HookFn] | HookFn | None]

class ABCOrAndMeta(type):
    @classmethod
    def __and__(cls: META, other: META) -> META: ...
    @classmethod
    def merge(cls: META, other: META) -> META: ...

class BaseMeta:
    __special_attrs__: _ClassVar[frozenset] = ...
    __is_inner_meta__: _ClassVar[bool] = ...
    recursive: _ClassVar[bool] = ...
    tag: _ClassVar[str | None] = ...
    tag_key: _ClassVar[str] = ...
    auto_assign_tags: _ClassVar[bool] = ...
    skip_defaults: _ClassVar[bool] = ...
    skip_if: _ClassVar[Condition | None] = ...
    skip_defaults_if: _ClassVar[Condition | None] = ...
    debug: _ClassVar[bool | int | str] = ...
    type_to_load_hook: _ClassVar[TypeToHook | None] = ...
    type_to_dump_hook: _ClassVar[TypeToHook | None] = ...
    pre_decoder: _ClassVar[PreDecoder] = ...
    dump_case: _ClassVar[KeyCase | str | None] = ...
    field_to_alias_dump: _ClassVar[
        typing.Mapping[str, str | typing.Sequence[str]] | None] = ...
    unsafe_parse_dataclass_in_union: _ClassVar[bool] = ...
    dump_date_time_as: _ClassVar[DateTimeTo | str | None] = ...
    assume_naive_datetime_tz: _ClassVar[tzinfo | None] = ...
    namedtuple_as_dict: _ClassVar[bool | None] = ...
    coerce_none_to_empty_str: _ClassVar[bool | None] = ...
    leaf_handling: _ClassVar[
        typing.Literal['exact', 'issubclass'] | None] = ...
    all_fields: _ClassVar[frozenset] = ...
    fields_to_merge: _ClassVar[frozenset] = ...

class AbstractMeta(BaseMeta):
    __special_attrs__: _ClassVar[frozenset] = ...
    __is_inner_meta__: _ClassVar[bool] = ...
    case: _ClassVar[KeyCase | str | None] = ...
    load_case: _ClassVar[KeyCase | str | None] = ...
    field_to_alias: _ClassVar[
        typing.Mapping[str, str | typing.Sequence[str]] | None] = ...
    field_to_alias_load: _ClassVar[
        typing.Mapping[str, str | typing.Sequence[str]] | None] = ...
    on_unknown_key: _ClassVar[KeyAction | None] = ...
    @classmethod
    def bind_to(
        cls, dataclass: type, create: bool = ..., is_default: bool = ...): ...

class AbstractEnvMeta(BaseMeta):
    __special_attrs__: _ClassVar[frozenset] = ...
    __is_inner_meta__: _ClassVar[bool] = ...
    env_file: _ClassVar[EnvFilePaths] = ...
    env_prefix: _ClassVar[str | None] = ...
    secrets_dir: _ClassVar[SecretsDirs] = ...
    load_case: _ClassVar[EnvKeyStrategy | str | None] = ...
    env_precedence: _ClassVar[EnvPrecedence | None] = ...
    field_to_env_load: _ClassVar[
        typing.Mapping[str, str | typing.Sequence[str]] | None] = ...
    @classmethod
    def bind_to(
        cls, env_class: type, create: bool = ..., is_default: bool = ...): ...

class _BaseHookRegistry:
    @classmethod
    def register_hook(cls, typ: type, func: Callable): ...
    @classmethod
    def get_hook(cls, typ: type) -> Callable | None: ...

class BaseLoadHook(_BaseHookRegistry):
    __HOOKS__: _ClassVar[dict[type, Callable]]

class BaseDumpHook(_BaseHookRegistry):
    __HOOKS__: _ClassVar[dict[type, Callable]]



from typing import Any

class Condition:

    op: str         # Operator
    val: Any        # Value
    t_or_f: bool    # Truthy or falsy
    _wrapped: bool  # True if wrapped in `SkipIf()`

    def __init__(self, operator: str, value: Any):
        ...

    def __str__(self):
        ...

    def evaluate(self, other) -> bool:
        ...


# Aliases for conditions
# noinspection PyPep8Naming
def EQ(value: Any) -> Condition:
    """Create a condition for equality (==)."""


# noinspection PyPep8Naming
def NE(value: Any) -> Condition:
    """Create a condition for inequality (!=)."""


# noinspection PyPep8Naming
def LT(value: Any) -> Condition:
    """Create a condition for less than (<)."""


# noinspection PyPep8Naming
def LE(value: Any) -> Condition:
    """Create a condition for less than or equal to (<=)."""


# noinspection PyPep8Naming
def GT(value: Any) -> Condition:
    """Create a condition for greater than (>)."""


# noinspection PyPep8Naming
def GE(value: Any) -> Condition:
    """Create a condition for greater than or equal to (>=)."""


# noinspection PyPep8Naming
def IS(value: Any) -> Condition:
    """Create a condition for identity (is)."""


# noinspection PyPep8Naming
def IS_NOT(value: Any) -> Condition:
    """Create a condition for non-identity (is not)."""


# noinspection PyPep8Naming
def IS_TRUTHY() -> Condition:
    """Create a "truthy" condition for evaluation (if <var>)."""


# noinspection PyPep8Naming
def IS_FALSY() -> Condition:
    """Create a "falsy" condition for evaluation (if not <var>)."""


# noinspection PyPep8Naming
def SkipIf(condition: Condition) -> Condition:
    ...


SkipIfNone: Condition



from __future__ import annotations

import collections.abc as abc
import dataclasses
from base64 import b64decode
from collections import defaultdict, deque
from dataclasses import MISSING, Field, is_dataclass
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Literal, NamedTuple, cast
from uuid import UUID

from ._bases import AbstractMeta, BaseLoadHook
from ._class_helper import (
    CLASS_TO_LOADER,
    DATACLASS_FIELD_TO_ALIAS_PATH_FOR_LOAD,
    resolve_dataclass_field_to_alias_for_load,
    set_class_loader,
)
from ._decorators import (
    process_patterned_date_time,
    setup_recursive_safe_function,
    setup_recursive_safe_function_for_generic,
)
from ._log import LOG
from ._meta_cache import get_meta
from ._models import LEAF_TYPES, Extras, TypeInfo
from ._models_date import UTC
from ._type_conv import (
    TRUTHY_VALUES,
    as_date,
    as_datetime,
    as_int,
    as_time,
    as_timedelta,
)
from ._type_def import (
    META,
    UNSET,
    DefFactory,
    JSONObject,
    NoneType,
    PyLiteralString,
    T,
)
from ._type_utils import create_new_class, is_subclass_safe

# noinspection PyUnresolvedReferences
from .constants import _HOOKS, CATCH_ALL, PACKAGE_NAME, PY311_OR_ABOVE, TAG
from .enums import KeyAction, KeyCase
from .errors import (
    JSONWizardError,
    MissingData,
    MissingFields,
    ParseError,
    UnknownKeysError,
)
from .utils._dataclass_compat import (
    SEEN_DEFAULT,
    dataclass_fields,
    dataclass_init_field_names,
    dataclass_init_fields,
    dataclass_kw_only_init_field_names,
    set_new_attribute,
)
from .utils._function_builder import FunctionBuilder
from .utils._object_path import safe_get
from .utils._string_conv import possible_json_keys
from .utils._typing_compat import (
    eval_forward_ref_if_needed,
    get_args,
    get_keys_for_typed_dict,
    get_origin_v2,
    is_annotated,
    is_typed_dict,
    is_typed_dict_type_qualifier,
    is_union,
)


class LoadMixin(BaseLoadHook):
    """
    This Mixin class derives its name from the eponymous `json.loads`
    function. Essentially it contains helper methods to convert JSON strings
    (or a Python dictionary object) to a `dataclass` which can often contain
    complex types such as lists, dicts, or even other dataclasses nested
    within it.

    Refer to the :class:`AbstractLoader` class for documentation on any of the
    implemented methods.

    """
    __slots__ = ()

    def __init_subclass__(cls, _setup_defaults=True, **kwargs):
        super().__init_subclass__(**kwargs)
        if _setup_defaults:
            setup_default_loader(cls)

    transform_json_field = None

    @staticmethod
    def load_fallback(tp: TypeInfo, extras: Extras):
        # identity: o
        return tp.v()

    @staticmethod
    def is_none(tp: TypeInfo, extras: Extras) -> str:
        return f'{tp.v()} is None'

    @classmethod
    def load_to_str(cls, tp: TypeInfo, extras: Extras):
        tn = tp.type_name(extras)
        o = tp.v()

        # str(v)
        if not extras['config'].coerce_none_to_empty_str or tp.in_optional:
            return f'{tn}({o})'

        # '' if v is None else str(v)
        default = "''" if tp.origin is str else f'{tn}()'
        return f'{default} if {cls.is_none(tp, extras)} else {tn}({o})'

    @staticmethod
    def load_to_int(tp: TypeInfo, extras: Extras):
        """
        Generate code to load a value into an integer field.

        Current logic to parse (an annotated) ``int`` returns:
          - ``v``          -->  ``v`` is an ``int`` or similarly annotated type.
          - ``int(v)``     -->  ``v`` is a ``str`` value of either a decimal
                                integer (e.g. ``'123'``) or a non-fractional
                                float value (e.g. ``42.0``).
          - ``as_int(v)``  -->  ``v`` is a non-fractional ``float``, or in case
                                of "less common" types / scenarios. Note that
                                empty strings and ``None`` (e.g. null values)
                                are not supported.

        """
        tn = tp.type_name(extras)
        o = tp.v()
        tp.ensure_in_locals(extras, as_int=as_int)

        return (
            f'{o} '
            f'if (t := {o}.__class__) is {tn} '
            f"else {tn}(f if '.' in {o} and (f := float({o})).is_integer() else {o}) "
            'if t is str '
            f'else as_int({o}, t, {tn})'
        )

        # TODO when `in_union`, we already know `o.__class__`
        #  is not `tn`, and we already have a variable `tp`.

    @staticmethod
    def load_to_float(tp: TypeInfo, extras: Extras):
        # alias: float(o)
        return tp.wrap_builtin(float, tp.v(), extras)

    @staticmethod
    def load_to_bool(tp: TypeInfo, extras: Extras):
        o = tp.v()
        tp.ensure_in_locals(extras, __TRUTHY=TRUTHY_VALUES)

        return (f'{o}.lower() in __TRUTHY '
                f'if {o}.__class__ is str '
                f'else {o} == 1')

    @staticmethod
    def load_to_bytes(tp: TypeInfo, extras: Extras):
        tp.ensure_in_locals(extras, b64decode)
        o = tp.v()
        return (f'{o} if (t := {o}.__class__) is bytes '
                f'else bytes({o}) if t is bytearray '
                f'else b64decode({o})')

    @classmethod
    def load_to_bytearray(cls, tp: TypeInfo, extras: Extras):
        # micro-optimization: avoid copying when already a bytearray
        # return f'{o} if (t := {o}.__class__) is bytearray else bytearray({o} if t is bytes else b64decode({o}))'
        as_bytes = cls.load_to_bytes(tp, extras)
        return tp.wrap_builtin(bytearray, as_bytes, extras)

    @staticmethod
    def load_to_none(tp: TypeInfo, extras: Extras):
        return 'None'

    @staticmethod
    def load_to_enum(tp: TypeInfo, extras: Extras):
        # alias: enum_cls(o)
        return tp.wrap(tp.v(), extras)

    @staticmethod
    def load_to_uuid(tp: TypeInfo, extras: Extras):
        # alias: UUID(o)
        return tp.wrap_builtin(UUID, tp.v(), extras)

    @classmethod
    def load_to_iterable(cls, tp: TypeInfo, extras: Extras):
        v, v_next, i_next = tp.v_and_next()
        gorg = tp.origin

        # noinspection PyBroadException
        try:
            elem_type = tp.args[0]
        except:
            elem_type = Any

        string = cls.load_dispatcher_for_annotation(
            tp.replace(origin=elem_type, i=i_next, index=None, val_name=None), extras)

        if issubclass(gorg, set):
            start_char = '{'
            end_char = '}'
        elif issubclass(gorg, frozenset):
            start_char = 'frozenset(('
            end_char = '))'
        else:
            start_char = '['
            end_char = ']'

        result = f'{start_char}{string} for {v_next} in {v}{end_char}'

        return tp.wrap(result, extras)

    @classmethod
    def load_to_tuple(cls, tp: TypeInfo, extras: Extras):
        args = tp.args

        # Determine the code string for the annotation

        # Check if the `Tuple` appears in the variadic form
        #   i.e. Tuple[str, ...]
        if args:
            is_variadic = args[-1] is ...
        else:
            # Annotated without args, as simply `tuple`
            args = (Any, ...)
            is_variadic = True

        if is_variadic:
            # Logic that handles the variadic form of :class:`Tuple`'s,
            # i.e. ``Tuple[str, ...]``
            #
            # Per `PEP 484`_, only **one** required type is allowed before the
            # ``Ellipsis``. That is, ``Tuple[int, ...]`` is valid whereas
            # ``Tuple[int, str, ...]`` would be invalid. `See here`_ for more info.
            #
            # .. _PEP 484: https://www.python.org/dev/peps/pep-0484/
            # .. _See here: https://github.com/python/typing/issues/180
            v, v_next, i_next = tp.v_and_next()

            # Given `Tuple[T, ...]`, we only need the generated string for `T`
            string = cls.load_dispatcher_for_annotation(
                tp.replace(origin=args[0], i=i_next, index=None, val_name=None), extras)

            result = f'[{string} for {v_next} in {v}]'

            # Wrap because we need to create a tuple from list comprehension
            force_wrap = True
        else:
            string = ', '.join([
                str(cls.load_dispatcher_for_annotation(
                    tp.replace(origin=arg, index=k, val_name=None),
                    extras))
                for k, arg in enumerate(args)])

            result = f'({string}, )'

            force_wrap = False

        return tp.wrap(result, extras, force=force_wrap)

    @classmethod
    def load_to_named_tuple(cls, tp: TypeInfo, extras: Extras):
        nt_tp = cast(NamedTuple, tp.origin)
        if nt_tp._field_defaults:  # has optionals
            return cls._load_to_named_tuple_fn(tp, extras)

        fields_in_order = nt_tp._fields  # field names in order
        ann = nt_tp.__annotations__

        if extras['config'].namedtuple_as_dict:
            values_in_order = tuple(
                str(
                    cls.load_dispatcher_for_annotation(
                        tp.replace(origin=ann.get(name, Any), index=repr(name)),
                        extras,
                    )
                )
                for name in fields_in_order
            )
        else:
            values_in_order = tuple(
                str(
                    cls.load_dispatcher_for_annotation(
                        tp.replace(origin=ann.get(name, Any), index=i),
                        extras,
                    )
                )
                for i, name in enumerate(fields_in_order)
            )

        params = ', '.join(values_in_order)
        return tp.wrap(params, extras)

    @classmethod
    @setup_recursive_safe_function(per_class_cache=True)
    def _load_to_named_tuple_fn(cls, tp: TypeInfo, extras: Extras):
        fn_gen = extras['fn_gen']
        _locals = extras['locals']

        nt_tp = _locals['cls'] = cast(NamedTuple, tp.origin)
        fields_in_order = nt_tp._fields  # field names in order
        field_to_default = nt_tp._field_defaults
        ann = nt_tp.__annotations__

        req_field_to_value = {}
        opt_field_to_value = {}

        all_optionals = len(field_to_default) == len(fields_in_order)
        v = tp.v_for_def()

        if extras['config'].namedtuple_as_dict:
            i_next = tp.i + 1
            v_next = f'{tp.prefix}{i_next}'

            for name in fields_in_order:
                field_tp = ann.get(name, Any)
                if name in field_to_default:
                    _locals[f'_dflt_{name}'] = field_to_default[name]
                    new_tp = tp.replace(origin=field_tp, i=i_next,
                                        index=None, val_name=None)
                    value = cls.load_dispatcher_for_annotation(new_tp, extras)
                    opt_field_to_value[name] = value
                else:
                    new_tp = tp.replace(origin=field_tp, index=repr(name))
                    value = cls.load_dispatcher_for_annotation(new_tp, extras)
                    req_field_to_value[f'__{name}'] = value

            req_args = ', '.join(req_field_to_value)
            opt_args = ', '.join(f'__{f}' for f in opt_field_to_value)

            if all_optionals:  # NamedTuple has no required fields
                ret_value_with_input = f'return cls({opt_args})'
            else:
                ret_value_with_input = f'return cls({req_args}, {opt_args})'
                for name, value in req_field_to_value.items():
                    fn_gen.add_line(f'{name} = {value}')

            # it's guaranteed the NamedTuple has at least one default field
            with fn_gen.if_(f'not {v}'):
                fn_gen.add_line('return cls()')

            for name, value in opt_field_to_value.items():
                with fn_gen.if_(f'({v_next} := {v}.get({name!r}, MISSING)) is MISSING'):
                    fn_gen.add_line(f'__{name} = _dflt_{name}')
                with fn_gen.else_():
                    fn_gen.add_line(f'__{name} = {value}')

            fn_gen.add_line(ret_value_with_input)

        else:  # list mode
            for i, name in enumerate(fields_in_order):
                field_tp = ann.get(name, Any)
                value = cls.load_dispatcher_for_annotation(
                    tp.replace(origin=field_tp, index=i), extras)

                if name in field_to_default:
                    opt_field_to_value[name] = value
                else:
                    req_field_to_value[f'__{name}'] = value

            req_args = ', '.join(req_field_to_value)
            opt_fields_start_i = len(req_field_to_value)

            if all_optionals:  # NamedTuple has no required fields
                len_condition = 'n'
                ret_value_with_input = 'return cls(*args)'
            else:
                len_condition = f'n > {opt_fields_start_i}'
                ret_value_with_input = f'return cls({req_args}, *args)'

                for name, value in req_field_to_value.items():
                    fn_gen.add_line(f'{name} = {value}')

            # it's guaranteed the NamedTuple has at least one default field
            fn_gen.add_line(f'n = len({v})')

            with fn_gen.if_(len_condition):
                opt_values = list(opt_field_to_value.values())
                fn_gen.add_line(f'args = [{opt_values.pop(0)}]')

                for i, value in enumerate(opt_values, start=opt_fields_start_i + 1):
                    with fn_gen.if_(f'n > {i}'):
                        fn_gen.add_line(f'args.append({value})')

                fn_gen.add_line(ret_value_with_input)

            fn_gen.add_line(f'return cls({req_args})')

    @classmethod
    def load_to_named_tuple_untyped(cls, tp: TypeInfo, extras: Extras):
        # Check if input object is `dict` or `list`.
        #
        # Assuming `Point` is a `namedtuple`, this performs
        # the equivalent logic as:
        #   Point(**x) if isinstance(x, dict) else Point(*x)
        #
        # star, dbl_star = tp.multi_wrap(extras, 'nt_', f'*{v}', f'**{v}')

        v = tp.v()

        if extras['config'].namedtuple_as_dict:
            return tp.wrap(f'**{v}', extras, prefix='nt_')

        def raise_():
            raise TypeError('Expected list/tuple for NamedTuple field') from None

        tp.ensure_in_locals(extras, raise_=raise_)

        star = tp.wrap(f'*{v}', extras, prefix='nt_')
        return f'{star} if (t := type({v})) is list or t is tuple else raise_()'

    @classmethod
    def _build_dict_comp(cls, tp, v, i_next, k_next, v_next, kt, vt, extras):
        tp_k_next = tp.replace(origin=kt, i=i_next, prefix='k', index=None, val_name=None)
        string_k = cls.load_dispatcher_for_annotation(tp_k_next, extras)

        tp_v_next = tp.replace(origin=vt, i=i_next, prefix='v', index=None, val_name=None)
        string_v = cls.load_dispatcher_for_annotation(tp_v_next, extras)

        return f'{{{string_k}: {string_v} for {k_next}, {v_next} in {v}.items()}}'

    @classmethod
    def load_to_dict(cls, tp: TypeInfo, extras: Extras):
        v, k_next, v_next, i_next = tp.v_and_next_k_v()

        try:
            kt, vt = tp.args
        except ValueError:
            # Annotated without two arguments,
            # e.g. like `dict[str]` or `dict`
            kt = vt = Any

        result = cls._build_dict_comp(
            tp, v, i_next, k_next, v_next, kt, vt, extras)

        return tp.wrap(result, extras)

    @classmethod
    def load_to_defaultdict(cls, tp: TypeInfo, extras: Extras):
        v, k_next, v_next, i_next = tp.v_and_next_k_v()
        default_factory: DefFactory | None

        try:
            kt, vt = tp.args
            default_factory = getattr(vt, '__origin__', vt)
        except ValueError:
            # Annotated without two arguments,
            # e.g. like `defaultdict[str]` or `defaultdict`
            kt = vt = Any
            default_factory = NoneType

        result = cls._build_dict_comp(
            tp, v, i_next, k_next, v_next, kt, vt, extras)

        return tp.wrap_dd(default_factory, result, extras)

    @classmethod
    def load_to_typed_dict(cls, tp: TypeInfo, extras: Extras):
        req_keys, opt_keys = get_keys_for_typed_dict(tp.origin)
        if opt_keys:  # has optionals
            return cls._load_to_typed_dict_fn(tp, extras)

        ann = tp.origin.__annotations__

        dict_body = ', '.join(
            f"""{name!r}: {
            cls.load_dispatcher_for_annotation(
                tp.replace(origin=ann.get(name, Any), index=repr(name)),
                extras,
            )
            }"""
            for name in req_keys
        )

        return f'{{{dict_body}}}'

    @classmethod
    @setup_recursive_safe_function
    def _load_to_typed_dict_fn(cls, tp: TypeInfo, extras: Extras):
        fn_gen = extras['fn_gen']

        req_keys, opt_keys = get_keys_for_typed_dict(tp.origin)

        result_list = []

        v = tp.v_for_def()
        # TODO set __annotations__?
        td_annotations = tp.origin.__annotations__

        # Set required keys for the `TypedDict`
        for k in req_keys:
            field_tp = td_annotations[k]
            field_name = repr(k)
            string = cls.load_dispatcher_for_annotation(
                tp.replace(origin=field_tp,
                           index=field_name,
                           val_name=None), extras)

            result_list.append(f'{field_name}: {string}')

        with fn_gen.try_():
            fn_gen.add_lines('result = {',
                             *(f'  {r},' for r in result_list),
                             '}')

            # Set optional keys for the `TypedDict` (if they exist)
            next_i = tp.i + 1
            new_tp = tp.replace(i=next_i, index=None, val_name=None)
            v_next = new_tp.v()

            for k in opt_keys:
                field_tp = td_annotations[k]
                field_name = repr(k)
                string = cls.load_dispatcher_for_annotation(
                    new_tp.replace(origin=field_tp), extras)
                with fn_gen.if_(f'({v_next} := {v}.get({field_name}, MISSING)) is not MISSING'):
                    fn_gen.add_line(f'result[{field_name}] = {string}')
            fn_gen.add_line('return result')

        with fn_gen.except_(Exception, 'e'):
            with fn_gen.if_('type(e) is KeyError'):
                fn_gen.add_line('name = e.args[0]; e = KeyError(f"Missing required key: {name!r}")')
            with fn_gen.elif_(f'not isinstance({v}, dict)'):
                fn_gen.add_line('e = TypeError("Incorrect type for object")')
            fn_gen.add_line(f"raise ParseError(e, {v}, {{}}, 'load') from None")

    @classmethod
    @setup_recursive_safe_function_for_generic(per_class_cache=True)
    def load_to_union(cls, tp: TypeInfo, extras: Extras):
        fn_gen = extras['fn_gen']
        config = extras['config']
        actual_cls = extras['cls']

        tag_key = config.tag_key or TAG
        auto_assign_tags = config.auto_assign_tags
        leaf_handling_as_subclass = config.leaf_handling == 'issubclass'

        args = tp.args
        in_optional = NoneType in args

        _locals = extras['locals']
        _locals['fields'] = args
        _locals['tag_key'] = tag_key

        dataclass_tag_to_lines: dict[str, list] = {}
        has_dataclass = any(is_dataclass(a) for a in args)

        i = tp.i
        v = tp.v_for_def()

        # TODO:
        #   Union handling here assumes `i == 1` (EnvWizard). If
        #   reused for multiple Union fields, cache/function-name
        #   collisions are possible.
        # noinspection PyUnboundLocalVariable
        if (has_dataclass
            and (pre_decoder := config.pre_decoder) is not None
            and (new_v := pre_decoder(cls, dict, tp, extras).v()) != v):
            current_v = v
            tp = tp.replace(i=i + 1)

            i = tp.i
            v = tp.v_for_def()

            with fn_gen.try_():
                fn_gen.add_line(f'{v} = {new_v}')
            with fn_gen.except_(Exception):
                fn_gen.add_line(f'{v} = {current_v}')

        type_checks = []
        try_parse_at_end = []

        for possible_tp in args:

            possible_tp = eval_forward_ref_if_needed(possible_tp, actual_cls)

            tp_new = TypeInfo(possible_tp, i=i)
            tp_new.in_optional = in_optional

            if possible_tp is NoneType:
                with fn_gen.if_(cls.is_none(tp, extras)):
                    fn_gen.add_line('return None')
                continue

            if has_dataclass and is_dataclass(possible_tp):
                # we see a dataclass in `Union` declaration
                meta = get_meta(possible_tp)
                tag = meta.tag
                assign_tags_to_cls = auto_assign_tags or meta.auto_assign_tags
                cls_name = possible_tp.__name__

                if assign_tags_to_cls and not tag:
                    tag = cls_name
                    # We don't want to mutate the base Meta class here
                    if meta is AbstractMeta:
                        from ._meta_cache import create_meta
                        create_meta(possible_tp, cls_name, tag=tag)
                    else:
                        meta.tag = cls_name

                if tag:
                    string = cls.load_dispatcher_for_annotation(tp_new, extras)

                    dataclass_tag_to_lines[tag] = [
                        f'if tag == {tag!r}:',
                        f'  return {string}'
                    ]
                    continue

                elif not config.unsafe_parse_dataclass_in_union:
                    e = ValueError('Cannot parse dataclass types in a Union without '
                                   'one of the following `Meta` settings:\n\n'
                                   '  * `auto_assign_tags = True`\n'
                                   f'    - Set on class `{extras["cls_name"]}`.\n\n'
                                   f'  * `tag = "{cls_name}"`\n'
                                   f'    - Set on class `{possible_tp.__qualname__}`.\n\n'
                                   '  * `unsafe_parse_dataclass_in_union = True`\n'
                                   f'    - Set on class `{extras["cls_name"]}`\n\n'
                                   'For more information, refer to:\n'
                                   '  https://dcw.ritviknag.com/en/latest/common_use_cases/dataclasses_in_union_types.html')
                    raise e from None

            string = cls.load_dispatcher_for_annotation(tp_new, extras)

            try_parse_lines = [
                'try:',
                f'  return {string}',
                'except Exception:',
                '  pass',
            ]

            if (possible_tp in LEAF_TYPES or (
                leaf_handling_as_subclass
                and is_subclass_safe(
                get_origin_v2(possible_tp), LEAF_TYPES)
            )):

                # TODO disable for dataclasses

                tn = tp_new.type_name(extras)
                type_checks.extend([
                    f'if tp is {tn}:',
                    f'  return {v}'
                ])
                list_to_add = try_parse_at_end
            else:
                list_to_add = type_checks

            list_to_add.extend(try_parse_lines)

        if dataclass_tag_to_lines:

            with fn_gen.try_():
                fn_gen.add_line(f'tag = {v}[tag_key]')

            with fn_gen.except_(Exception):
                fn_gen.add_line('pass')

            with fn_gen.else_():

                for lines in dataclass_tag_to_lines.values():
                    fn_gen.add_lines(*lines)
                fn_gen.add_line(
                    "raise ParseError("
                    "TypeError('Object with tag was not in any of Union types'),"
                    f"{v},fields,'load',"
                    "input_tag=tag,"
                    "tag_key=tag_key,"
                    f"valid_tags={list(dataclass_tag_to_lines)})"
                )

        fn_gen.add_line(f'tp = type({v})')

        if type_checks:
            fn_gen.add_lines(*type_checks)

        if try_parse_at_end:
            fn_gen.add_lines(*try_parse_at_end)

        # Invalid type for Union
        fn_gen.add_line("raise ParseError("
                        "TypeError('Object was not in any of Union types'),"
                        f"{v},fields,'load',"
                        "tag_key=tag_key"
                        ")")

    @staticmethod
    @setup_recursive_safe_function_for_generic
    def load_to_literal(tp: TypeInfo, extras: Extras):
        fn_gen = extras['fn_gen']

        v = tp.v_for_def()

        _locals = extras['locals']
        _locals['fields'] = frozenset(tp.args)

        with fn_gen.if_(f'{v} in fields', comment=repr(tp.args)):
            fn_gen.add_line(f'return {v}')

        # No such Literal with the value of `o`
        fn_gen.add_line("e = ValueError('Value not in expected Literal values')")
        fn_gen.add_line(f"raise ParseError(e, {v}, fields, 'load', "
                        f'allowed_values=list(fields))')

        # TODO Checks for Literal equivalence, as mentioned here:
        #   https://www.python.org/dev/peps/pep-0586/#equivalence-of-two-literals

        # extras_cp['locals'][fields] = {
        #     a: type(a) for a in tp.args
        # }
        #
        # with fn_gen.function(fn_name, ['v1'], None, _locals):
        #
        #     with fn_gen.try_():
        #         with fn_gen.if_(f'type({tp.v()}) is {fields}[{tp.v()}]'):
        #             fn_gen.add_line('return v1')
        #
        #         # The value of `o` is in the ones defined for the Literal, but
        #         # also confirm the type matches the one defined for the Literal.
        #         fn_gen.add_line("e = TypeError('Value did not match expected type for the Literal')")
        #
        #         fn_gen.add_line('raise ParseError('
        #                         f'e, v1, {fields}, '
        #                         'have_type=type(v1), '
        #                         f'desired_type={fields}[v1], '
        #                         f'desired_value=next(v for v in {fields} if v == v1), '
        #                         f'allowed_values=list({fields})'
        #                         ')')
        #     with fn_gen.except_(KeyError):
        #         # No such Literal with the value of `o`
        #         fn_gen.add_line("e = ValueError('Value not in expected Literal values')")
        #         fn_gen.add_line('raise ParseError('
        #                         f'e, v1, {fields}, allowed_values=list({fields})'
        #                         f')')

    @staticmethod
    def load_to_decimal(tp: TypeInfo, extras: Extras):
        o = tp.v()
        s = f'str({o}) if {o}.__class__ is float else {o}'

        return tp.wrap_builtin(Decimal, s, extras)

    @staticmethod
    def load_to_path(tp: TypeInfo, extras: Extras):
        # alias: Path(o)
        return tp.wrap_builtin(Path, tp.v(), extras)

    @classmethod
    @process_patterned_date_time
    def load_to_date(cls, tp: TypeInfo, extras: Extras):
        return cls._load_to_date(tp, extras, date)

    @classmethod
    @process_patterned_date_time
    def load_to_datetime(cls, tp: TypeInfo, extras: Extras):
        return cls._load_to_date(tp, extras, datetime)

    @staticmethod
    @process_patterned_date_time
    def load_to_time(tp: TypeInfo, extras: Extras):
        o = tp.v()
        tn = tp.type_name(extras, bound=time)
        tp_time = cast('type[time]', tp.origin)

        __fromisoformat = f'__{tn}_fromisoformat'

        tp.ensure_in_locals(
            extras,
            __as_time=as_time,
            **{__fromisoformat: tp_time.fromisoformat}
        )

        if PY311_OR_ABOVE:
            _parse_iso_string = f'{__fromisoformat}({o})'
        else:  # pragma: no cover
            _parse_iso_string = f"{__fromisoformat}({o}.replace('Z', '+00:00', 1))"

        return (f'{_parse_iso_string} if {o}.__class__ is str '
                f'else __as_time({o}, {tn})')

    @staticmethod
    def _load_to_date(tp: TypeInfo, extras: Extras,
                      cls: type[date] | type[datetime]):
        o = tp.v()
        tn = tp.type_name(extras, bound=cls)
        tp_date_or_datetime = cast('type[date]', tp.origin)

        _fromisoformat = f'__{tn}_fromisoformat'

        name_to_func = {
            _fromisoformat: tp_date_or_datetime.fromisoformat,
        }

        if cls is datetime:  # datetime or a subclass
            _fromtimestamp = f'__{tn}_fromtimestamp'
            name_to_func[_fromtimestamp] = tp_date_or_datetime.fromtimestamp
            _as_func = '__as_datetime'
            name_to_func[_as_func] = as_datetime
            _date_part = _opt_cls = ''

        else:  # date or a subclass
            _fromtimestamp = '__datetime_fromtimestamp'
            name_to_func[_fromtimestamp] = datetime.fromtimestamp
            _as_func = '__as_date'
            name_to_func[_as_func] = as_date
            _date_part = '.date()'
            _opt_cls = f', {tn}'

        tp.ensure_in_locals(extras, UTC=UTC, **name_to_func)

        if PY311_OR_ABOVE:
            _parse_iso_string = f'{_fromisoformat}({o})'
        else:  # pragma: no cover
            _parse_iso_string = f"{_fromisoformat}({o}.replace('Z', '+00:00', 1))"

        return (f'({_fromtimestamp}(int({o}), UTC){_date_part} if {o}.isdigit() '
                f'else {_parse_iso_string}) if {o}.__class__ is str '
                f'else {_as_func}({o}, {_fromtimestamp}, UTC{_opt_cls})')

    @staticmethod
    def load_to_timedelta(tp: TypeInfo, extras: Extras):
        # alias: as_timedelta
        tn = tp.type_name(extras, bound=timedelta)
        tp.ensure_in_locals(extras, as_timedelta)

        return f'as_timedelta({tp.v()}, {tn})'

    @staticmethod
    @setup_recursive_safe_function(
        fn_name=f'__{PACKAGE_NAME}_from_dict_{{cls_name}}__')
    def load_to_dataclass(tp: TypeInfo, extras: Extras):
        load_func_for_dataclass(tp.origin, extras)

    @classmethod
    def load_dispatcher_for_annotation(cls,
                                       tp,
                                       extras):

        hooks = cls.__HOOKS__
        config = extras['config']
        pre_decoder = config.pre_decoder
        type_hooks = config.type_to_load_hook
        leaf_handling_as_subclass = config.leaf_handling == 'issubclass'

        # type_ann = tp.origin
        type_ann = eval_forward_ref_if_needed(tp.origin, extras['cls'])

        origin = get_origin_v2(type_ann)
        name = getattr(origin, '__name__', origin)
        args = None

        container_tp = None

        if is_annotated(type_ann):
            # Given `Annotated[T, ...]`, we only need `T`
            type_ann, *field_extras = get_args(type_ann)
            type_ann = eval_forward_ref_if_needed(type_ann, extras['cls'])
            origin = get_origin_v2(type_ann)
            name = getattr(origin, '__name__', origin)

            # Check for Custom Patterns for date / time / datetime
            for extra in field_extras:
                if getattr(extra, '__dcw_pattern__', False):
                    extras['pattern'] = extra

        elif is_typed_dict_type_qualifier(origin):
            # Given `Required[T]` or `NotRequired[T]`, we only need `T`
            # Loop because they can be nested `ReadOnly[NotRequired[...]]
            while is_typed_dict_type_qualifier(origin):
                type_ann = get_args(type_ann)[0]
                origin = get_origin_v2(type_ann)
                name = getattr(origin, '__name__', origin)

        # TypeAliasType: Type aliases are created through
        # the `type` statement
        if (value := getattr(origin, '__value__', None)) is not None:
            type_ann = value
            origin = get_origin_v2(type_ann)
            name = getattr(origin, '__name__', origin)

        # `LiteralString` enforces stricter rules at
        # type-checking but behaves like `str` at runtime.
        # TODO maybe add `load_to_literal_string`
        if origin is PyLiteralString:
            load_hook = cls.load_to_str
            origin = str
            name = 'str'

        # -> Atomic, immutable types which don't require
        #    any iterative / recursive handling.
        elif origin in LEAF_TYPES or (
            leaf_handling_as_subclass
            and is_subclass_safe(origin, LEAF_TYPES)):
            load_hook = hooks.get(origin)

        elif (type_hooks is not None
              and (hook_info := type_hooks.get(origin)) is not None):
            mode, load_hook = hook_info

            if mode == 'runtime':
                fn_name, = tp.ensure_in_locals(extras, load_hook)
                return f'{fn_name}({tp.v()})'

            try:
                args = get_args(type_ann)
            except ValueError:
                args = Any,

        elif (load_hook := hooks.get(origin)) is not None:
            try:
                args = get_args(type_ann)
            except ValueError:
                args = Any,

        # -> Union[x]
        elif is_union(origin):
            load_hook = cls.load_to_union
            args = get_args(type_ann)

            # Special case for Optional[x], which is actually Union[x, None]
            if len(args) == 2 and NoneType in args:
                new_tp = tp.replace(origin=args[0], args=None, name=None, val_name=None)
                new_tp.in_optional = True

                string = cls.load_dispatcher_for_annotation(new_tp, extras)

                return f'None if {cls.is_none(tp, extras)} else {string}'

        # -> Literal[X, Y, ...]
        elif origin is Literal:
            load_hook = cls.load_to_literal
            args = get_args(type_ann)

        # https://stackoverflow.com/questions/76520264/dataclasswizard-after-upgrading-to-python3-11-is-not-working-as-expected
        elif origin is Any:
            load_hook = cls.load_fallback

        elif is_subclass_safe(origin, tuple) and hasattr(origin, '_fields'):
            container_tp = dict if config.namedtuple_as_dict else tuple
            if getattr(origin, '__annotations__', None):
                # Annotated as a `typing.NamedTuple` subtype
                load_hook = cls.load_to_named_tuple
            else:
                # Annotated as a `collections.namedtuple` subtype
                load_hook = cls.load_to_named_tuple_untyped

        elif is_typed_dict(origin):
            container_tp = dict
            load_hook = cls.load_to_typed_dict

        elif is_dataclass(origin):
            container_tp = dict
            # return a dynamically generated `fromdict`
            # for the `cls` (base_type)
            load_hook = cls.load_to_dataclass

        elif is_subclass_safe(origin, Enum):
            load_hook = cls.load_to_enum

        elif origin in (abc.Sequence, abc.MutableSequence, abc.Collection):
            if origin is abc.Sequence:
                load_hook = cls.load_to_tuple
                # desired (non-generic) origin type
                name = 'tuple'
                origin = tuple
                # Re-map type arguments to variadic tuple format,
                # e.g. `Sequence[int]` -> `tuple[int, ...]`
                try:
                    args = (get_args(type_ann)[0], ...)
                except (IndexError, ValueError):
                    args = Any,
            else:
                load_hook = cls.load_to_iterable
                # desired (non-generic) origin type
                name = 'list'
                origin = list
                # Get type arguments, e.g. `Sequence[int]` -> `int`
                try:
                    args = get_args(type_ann)
                except ValueError:
                    args = Any,

        elif getattr(origin, '__dcw_pattern__', False):
            load_hook = origin.load_to_pattern

        else:

            # TODO everything should use `get_origin_v2`
            try:
                args = get_args(type_ann)
            except ValueError:
                args = Any,

        if load_hook is None:
            # TODO END
            for t in hooks:
                if (not leaf_handling_as_subclass) and (t in LEAF_TYPES):
                    continue
                if issubclass(origin, t):
                    container_tp = t
                    load_hook = hooks[t]
                    break

        tp.origin = origin
        tp.args = args
        tp.name = name

        if container_tp is None:
            container_tp = origin

        if pre_decoder is not None:
            tp = pre_decoder(cls, container_tp, tp, extras)

        if load_hook is not None:
            result = load_hook(tp, extras)
            return result

        # No matching hook is found for the type.
        # TODO do we want to add a `Meta` field to not raise
        #  an error but perform a default action?
        err = TypeError('Provided type is not currently supported.')
        pe = ParseError(
            err, origin, type_ann, 'load',
            resolution=f'Register a load hook for {ParseError.name(origin)} '
                       f'(`register_type` / `Meta.type_to_load_hook`).',
            unsupported_type=origin
        )
        raise pe from None


def get_default_load_hooks(loader=LoadMixin):
    return {
        # Simple types
        str: loader.load_to_str,
        float: loader.load_to_float,
        bool: loader.load_to_bool,
        int: loader.load_to_int,
        bytes: loader.load_to_bytes,
        bytearray: loader.load_to_bytearray,
        NoneType: loader.load_to_none,
        # Complex types
        UUID: loader.load_to_uuid,
        set: loader.load_to_iterable,
        frozenset: loader.load_to_iterable,
        deque: loader.load_to_iterable,
        list: loader.load_to_iterable,
        tuple: loader.load_to_tuple,
        defaultdict: loader.load_to_defaultdict,
        dict: loader.load_to_dict,
        Decimal: loader.load_to_decimal,
        Path: loader.load_to_path,
        # Dates and times
        datetime: loader.load_to_datetime,
        time: loader.load_to_time,
        date: loader.load_to_date,
        timedelta: loader.load_to_timedelta,
    }


def setup_default_loader(cls=LoadMixin):
    """
    Set up the default type hooks to use when converting `str` (json) or a
    Python `dict` object to a `dataclass` instance.

    Note: `cls` must be :class:`LoadMixIn` or a subclass of it.
    """
    if '__HOOKS__' in cls.__dict__:
        return

    parent_hooks = getattr(cls, '__HOOKS__', None)

    hooks = get_default_load_hooks(cls)
    if parent_hooks:
        hooks |= parent_hooks  # parent / custom wins

    cls.__HOOKS__ = hooks


def check_and_raise_missing_fields(
    _locals, o, cls,
    fields: tuple[Field, ...] | None,
    **kwargs,
):
    if fields is None:  # `typing.NamedTuple` or `collections.namedtuple`
        nt_tp = cast(NamedTuple, cls)
        field_to_default = nt_tp._field_defaults
        field_names = nt_tp._fields

        fields = tuple([
            dataclasses.field(
                default=field_to_default.get(field, MISSING),
            )
            for field in field_names])

        for field, name in zip(fields, field_names):
            field.name = name

        missing_fields = [f for f in field_names
                          if f'__{f}' not in _locals
                          and f not in field_to_default]

        missing_keys = None

    else:
        missing_fields = [f.name for f in fields
                          if f.init
                          and f'__{f.name}' not in _locals
                          and (f.default is MISSING
                               and f.default_factory is MISSING)]

        missing_keys = [resolve_dataclass_field_to_alias_for_load(cls).get(field, [field])[0]
                        for field in missing_fields]

    raise MissingFields(
        None, o, cls, fields, None, missing_fields,
        missing_keys, **kwargs,
    ) from None


def load_func_for_dataclass(
    cls: type,
    extras: Extras | None = None,
    loader_cls=LoadMixin,
    base_meta_cls: type = AbstractMeta,
) -> Callable[[JSONObject], T] | None:
    # Tuple describing the fields of this dataclass.
    fields = dataclass_fields(cls)

    cls_init_fields = dataclass_init_fields(cls, True)
    cls_init_field_names = dataclass_init_field_names(cls)
    cls_init_kw_only_field_names = dataclass_kw_only_init_field_names(cls)

    # Get the loader for the class, or create a new one as needed.
    cls_loader = get_loader(cls, base_cls=loader_cls)

    cls_name = cls.__name__

    fn_name = f'__{PACKAGE_NAME}_from_dict_{cls_name}__'

    # Get the meta config for the class, or the default config otherwise.
    meta = get_meta(cls, base_meta_cls)

    if extras is None:  # we are being run for the main dataclass
        is_main_class = True

        # If the `recursive` flag is enabled and a Meta config is provided,
        # apply the Meta recursively to any nested classes.
        #
        # Else, just use the base `AbstractMeta`.
        config: META = meta if meta.recursive else base_meta_cls

        # Initialize the FuncBuilder
        fn_gen = FunctionBuilder()

        new_locals = {
            'cls': cls,
            'fields': fields,
        }

        extras: Extras = {
            'config': config,
            'cls': cls,
            'cls_name': cls_name,
            'locals': new_locals,
            'recursion_guard': {cls: fn_name},
            'fn_gen': fn_gen,
        }

        _globals = {
            'MISSING': MISSING,
            'ParseError': ParseError,
            'raise_missing_fields': check_and_raise_missing_fields,
            're_raise': re_raise,
        }

    # we are being run for a nested dataclass
    else:
        is_main_class = False

        # config for nested dataclasses
        config = extras['config']

        # Initialize the FuncBuilder
        fn_gen = extras['fn_gen']

        if config is not base_meta_cls:
            # we want to apply the meta config from the main dataclass
            # recursively.
            meta = meta | config
            meta.bind_to(cls, is_default=False)

        new_locals = extras['locals']
        new_locals['fields'] = fields

        # TODO need a way to auto-magically do this
        extras['cls'] = cls
        extras['cls_name'] = cls_name

        # Added for a `EnvWizard` main class, which doesn't set this in globals
        fn_gen.globals.setdefault('raise_missing_fields', check_and_raise_missing_fields)

    key_case: KeyCase | None = cls_loader.transform_json_field
    auto_key_case = key_case is KeyCase.AUTO

    field_to_aliases = resolve_dataclass_field_to_alias_for_load(cls)
    check_aliases = True if field_to_aliases else False

    field_to_paths = DATACLASS_FIELD_TO_ALIAS_PATH_FOR_LOAD[cls]
    has_alias_paths = True if field_to_paths else False

    # FIXME get from functions instead
    has_defaults = SEEN_DEFAULT[cls]

    # Fix for using `auto_assign_tags` and `on_unknown_key='RAISE'` together
    # See https://github.com/rnag/dataclass-wizard/issues/137
    has_tag_assigned = meta.tag is not None
    if (has_tag_assigned and
        # Ensure `tag_key` isn't a dataclass field,
        # to avoid issues with our logic.
        # See https://github.com/rnag/dataclass-wizard/issues/148
        meta.tag_key not in cls_init_field_names):
        expect_tag_as_unknown_key = True
    else:
        expect_tag_as_unknown_key = False

    on_unknown_key = meta.on_unknown_key

    catch_all_field: str | None = field_to_aliases.pop(CATCH_ALL, None)
    has_catch_all = catch_all_field is not None

    if has_catch_all:
        pre_assign = 'i+=1; '
        catch_all_field_stripped = catch_all_field.rstrip('?')
        catch_all_idx = cls_init_field_names.index(catch_all_field_stripped)
        # remove catch all field from list, so we don't iterate over it
        del cls_init_fields[catch_all_idx]
    else:
        pre_assign = ''
        catch_all_field_stripped = catch_all_idx = None

    if on_unknown_key is not None:
        should_raise = on_unknown_key is KeyAction.RAISE
        should_warn = on_unknown_key is KeyAction.WARN
        if should_warn or should_raise:
            pre_assign = 'i+=1; '
            set_aliases = True
        else:
            set_aliases = has_catch_all
    else:
        should_raise = should_warn = None
        set_aliases = has_catch_all

    if set_aliases:
        if expect_tag_as_unknown_key:
            # add an alias for the tag key, so we don't
            # capture or raise an error when we see it
            aliases = {meta.tag_key}
        else:
            aliases = set()

        new_locals['aliases'] = aliases
    else:
        aliases = None

    if has_alias_paths:
        new_locals['safe_get'] = safe_get

    with fn_gen.function(fn_name, ['o'], MISSING, new_locals):

        if (_pre_from_dict := getattr(cls, '_pre_from_dict', None)) is not None:
            new_locals['__pre_from_dict__'] = _pre_from_dict
            fn_gen.add_line('o = __pre_from_dict__(o)')

        # Need to create a separate dictionary to copy over the constructor
        # args, as we don't want to mutate the original dictionary object.
        if has_defaults:
            fn_gen.add_line('init_kwargs = {}')
        if pre_assign:
            fn_gen.add_line('i = 0')

        args = []
        kwargs = []

        if cls_init_fields:

            with fn_gen.try_():

                if expect_tag_as_unknown_key and pre_assign:
                    with fn_gen.if_(f'{meta.tag_key!r} in o'):
                        fn_gen.add_line('i+=1')

                val = 'v1'
                _val_is_found = f'{val} is not MISSING'
                for i, f in enumerate(cls_init_fields):
                    name = f.name
                    var = f'__{name}'
                    has_default = has_defaults and (
                        f.default is not MISSING
                        or f.default_factory is not MISSING
                    )
                    val_is_found = _val_is_found

                    if (check_aliases
                        and (_aliases := field_to_aliases.get(name)) is not None):

                        if len(_aliases) == 1:
                            alias = _aliases[0]

                            if set_aliases:
                                aliases.add(alias)

                            f_assign = f'field={name!r}; {val}=o.get({alias!r}, MISSING)'
                        else:
                            f_assign = None

                            # add possible JSON keys
                            if set_aliases:
                                aliases.update(_aliases)

                            fn_gen.add_line(f'field={name!r}')
                            condition = [f'({val} := o.get({alias!r}, MISSING)) is not MISSING'
                                         for alias in _aliases]

                            val_is_found = '(' + '\n     or '.join(condition) + ')'

                    elif (has_alias_paths
                          and (paths := field_to_paths.get(name)) is not None):

                        if len(paths) == 1:
                            path = paths[0]

                            # add the first part (top-level key) of the path
                            if set_aliases:
                                aliases.add(path[0])

                            f_assign = f'field={name!r}; {val}=safe_get(o, {path!r}, {not has_default})'
                        else:
                            f_assign = None
                            fn_gen.add_line(f'field={name!r}')
                            condition = []
                            last_idx = len(paths) - 1
                            for k, path in enumerate(paths):

                                # add the first part (top-level key) of each path
                                if set_aliases:
                                    aliases.add(path[0])

                                if k == last_idx:
                                    condition.append(
                                        f'({val} := safe_get(o, {path!r}, {not has_default})) is not MISSING')
                                else:
                                    condition.append(
                                        f'({val} := safe_get(o, {path!r}, False)) is not MISSING')

                            val_is_found = '(' + '\n     or '.join(condition) + ')'

                        # TODO raise some useful message like (ex. on IndexError):
                        #       Field "my_str" of type tuple[float, str] in A2 has invalid value ['123']

                    elif key_case is None:

                        if set_aliases:
                            aliases.add(name)

                        f_assign = f'field={name!r}; {val}=o.get(field, MISSING)'

                    elif auto_key_case:
                        f_assign = None

                        _aliases = possible_json_keys(name)

                        if set_aliases:
                            # add field name itself
                            aliases.add(name)
                            # add possible JSON keys
                            aliases.update(_aliases)

                        fn_gen.add_line(f'field={name!r}')
                        condition = [f'({val} := o.get(field, MISSING)) is not MISSING']
                        for alias in _aliases:
                            condition.append(f'({val} := o.get({alias!r}, MISSING)) is not MISSING')

                        val_is_found = '(' + '\n     or '.join(condition) + ')'

                    else:
                        alias = key_case(name)

                        if set_aliases:
                            aliases.add(alias)

                        if alias != name:
                            field_to_aliases[name] = (alias,)

                        f_assign = f'field={name!r}; {val}=o.get({alias!r}, MISSING)'

                    string = generate_field_code(cls_loader, extras, f, i)

                    if f_assign is not None:
                        fn_gen.add_line(f_assign)

                    if has_default:
                        with fn_gen.if_(val_is_found):
                            fn_gen.add_line(f'{pre_assign}init_kwargs[field] = {string}')

                    else:
                        if name in cls_init_kw_only_field_names:
                            kwargs.append(f'{name}={var}')
                        else:
                            args.append(var)

                        with fn_gen.if_(val_is_found):
                            fn_gen.add_line(f'{pre_assign}{var} = {string}')

            # create a broad `except Exception` block, as we will be
            # re-raising all exception(s) as a custom `ParseError`.
            with fn_gen.except_(Exception, 'e', ParseError):
                fn_gen.add_line("re_raise(e, cls, o, fields, field, locals().get('v1'))")

        if has_catch_all:
            catch_all_def = '{k: o[k] for k in o if k not in aliases}'

            if catch_all_field.endswith('?'):  # Default value
                with fn_gen.if_('len(o) != i'):
                    fn_gen.add_line(f'init_kwargs[{catch_all_field_stripped!r}] = {catch_all_def}')
            else:
                var = f'__{catch_all_field_stripped}'
                fn_gen.add_line(f'{var} = {{}} if len(o) == i else {catch_all_def}')

                if catch_all_field_stripped in cls_init_kw_only_field_names:
                    kwargs.append(f'{catch_all_field_stripped}={var}')
                else:
                    args.insert(catch_all_idx, var)

        elif set_aliases:  # warn / raise on unknown key
            line = 'extra_keys = set(o) - aliases'

            with fn_gen.if_('len(o) != i'):
                fn_gen.add_line(line)
                if should_raise:
                    # Raise an error here (if needed)
                    new_locals['UnknownKeysError'] = UnknownKeysError
                    fn_gen.add_line("raise UnknownKeysError(extra_keys, o, cls, fields) from None")
                elif should_warn:
                    # Show a warning here
                    new_locals['LOG'] = LOG
                    fn_gen.add_line(r"LOG.warning('Found %d unknown keys %r not mapped to the dataclass schema.\n"
                                    r"  Class: %r\n  Dataclass fields: %r', "
                                    "len(extra_keys), extra_keys, "
                                    "cls.__qualname__, [f.name for f in fields])")

        # Now pass the arguments to the constructor method, and return
        # the new dataclass instance. If there are any missing fields,
        # we raise them here.

        if has_defaults:
            args.append('**init_kwargs')
        if kwargs:
            args.extend(kwargs)
        with fn_gen.try_():
            fn_gen.add_line(f'return cls({", ".join(args)})')
        with fn_gen.except_(UnboundLocalError):
            # raise `MissingFields`, as required dataclass fields
            # are not present in the input object `o`.
            fn_gen.add_line("raise_missing_fields(locals(), o, cls, fields)")

    # Save the load function for the main dataclass, so we don't need to run
    # this logic each time.
    if is_main_class:
        # noinspection PyUnboundLocalVariable
        functions = fn_gen.create_functions(_globals)

        cls_fromdict = functions[fn_name]

        # Check if the class has a `from_dict`, and it's
        # a class method bound to `fromdict`.
        if ((from_dict := getattr(cls, 'from_dict', None)) is not None
            and getattr(from_dict, '__func__', None) is fromdict):
            LOG.debug("setattr(%s, 'from_dict', %s)", cls_name, fn_name)
            # Marker reserved for future detection/debugging of specialized loaders.
            # setattr(cls_fromdict, _SPECIALIZED_FROM_DICT, True)
            # safe to specialize only when user didn't define it on cls
            set_new_attribute(cls, 'from_dict', cls_fromdict, force=True)

        set_new_attribute(
            cls, '__dataclass_wizard_from_dict__', cls_fromdict, force=True)
        LOG.debug(
            "setattr(%s, '__%s_from_dict__', %s)",
            cls_name, PACKAGE_NAME, fn_name)

        return cls_fromdict


def generate_field_code(cls_loader: LoadMixin,
                        extras: Extras,
                        field: Field,
                        field_i: int,
                        var_name=None) -> str | TypeInfo:
    cls = extras['cls']
    field_type = field.type = eval_forward_ref_if_needed(field.type, cls)

    try:
        return cls_loader.load_dispatcher_for_annotation(
            TypeInfo(field_type, field_i=field_i, val_name=var_name), extras
        )

    # except Exception as e:
    #     re_raise(e, cls, None, dataclass_init_fields(cls), field, None)
    except ParseError as pe:
        pe.class_name = cls
        # noinspection PyPropertyAccess
        pe.field_name = field.name
        raise pe from None


def re_raise(e, cls, o, fields, field, value):
    # If the object `o` is None, then raise an error with
    # the relevant info included.
    if o is None:
        raise MissingData(cls) from None

    # Check if the object `o` is some other type than what we expect -
    # for example, we could be passed in a `list` type instead.
    if not isinstance(o, dict):
        base_err = TypeError('Incorrect type for `from_dict()`')
        e = ParseError(base_err, o, dict, 'load', cls, desired_type=dict)

    add_fields = True
    if type(e) is not ParseError:
        if isinstance(e, JSONWizardError):
            add_fields = False
        else:
            tp = getattr(next((f for f in fields if f.name == field), None), 'type', Any)
            e = ParseError(e, value, tp, 'load')

    # We run into a parsing error while loading the field value;
    # Add additional info on the Exception object before re-raising it.
    #
    # First confirm these values are not already set by an
    # inner dataclass. If so, it likely makes it easier to
    # debug the cause. Note that this should already be
    # handled by the `setter` methods.
    if add_fields:
        e.class_name, e.fields, e.field_name, e.json_object = cls, fields, field, o
    else:
        e.class_name, e.field_name, e.json_object = cls, field, o

    # noinspection PyUnboundLocalVariable
    if (isinstance(e, ParseError)
        # `typing.NamedTuple` or `collections.namedtuple`
        and (origin := e.ann_type) is not None
        and is_subclass_safe(origin, tuple)
        and (_fields := getattr(origin, '_fields', None))):

        meta = get_meta(cls)
        nt_tp = cast(NamedTuple, origin)
        field_to_default = nt_tp._field_defaults
        num_req_fields = len(_fields) - len(field_to_default)

        e_cls = getattr(e.base_error, '__class__', None)

        if e_cls in (IndexError, KeyError, TypeError):
            # raise `MissingFields`, as required NamedTuple fields
            # are not present in the input object `o`.
            if isinstance(value, (list, tuple)):
                # noinspection PyUnboundLocalVariable
                _locals = {f'__{f}' for f in _fields[:len(value)]}
                num_req_fields_provided = min(len(value), num_req_fields)
            elif isinstance(value, dict):
                _locals = {f'__{f}' for f in _fields if f in value}
                num_req_fields_provided = len(
                    [f for f in _fields
                     if f in value and f not in field_to_default]
                )
            else:
                _locals = _fields
                num_req_fields_provided = num_req_fields

            if num_req_fields_provided < num_req_fields:
                check_and_raise_missing_fields(
                    _locals, value, origin, None,
                    **(
                        {'field': f'{ParseError.name(cls)}.{field}'}
                        if cls and field
                        else {}
                    ))

        if meta.namedtuple_as_dict:
            if e_cls is TypeError and type(value) is not dict:
                e.kwargs['resolution'] = (
                    'List/tuple input is not supported for NamedTuple fields in dict mode. '
                    'Pass a dict, or set Meta.namedtuple_as_dict = False.'
                )
                e.kwargs['unsupported_type'] = type(value)
        else:
            if e_cls is KeyError and type(value) is dict:
                e.kwargs['resolution'] = (
                    'Dict input is not supported for NamedTuple fields in list mode. '
                    'Pass a list/tuple, or set Meta.namedtuple_as_dict = True.'
                )
                e.kwargs['unsupported_type'] = dict

    raise e from None


def get_loader(class_or_instance=None,
               create=True,
               base_cls: T = LoadMixin) -> type[T]:
    """
    Get the loader for the class, using the following logic:

        * Return the class if it's already a sub-class of :class:`LoadMixin`
        * If `create` is enabled (which is the default), a new sub-class of
          :class:`LoadMixin` for the class will be generated and cached on the
          initial run.
        * Otherwise, we will return the base loader, :class:`LoadMixin`, which
          can potentially be shared by more than one dataclass.

    """
    # TODO
    try:
        return CLASS_TO_LOADER[class_or_instance]

    except KeyError:

        if hasattr(class_or_instance, _HOOKS):
            return set_class_loader(
                CLASS_TO_LOADER, class_or_instance, class_or_instance)

        elif create:
            cls_loader = create_new_class(class_or_instance, (base_cls, ))
            return set_class_loader(
                CLASS_TO_LOADER, class_or_instance, cls_loader)

        return set_class_loader(
            CLASS_TO_LOADER, class_or_instance, base_cls)


def fromdict(cls: type[T], d: JSONObject) -> T:
    """
    Converts a Python dictionary object to a dataclass instance.

    Iterates over each dataclass field recursively; lists, dicts, and nested
    dataclasses will likewise be initialized as expected.

    When directly invoking this function, an optional Meta configuration for
    the dataclass can be specified via ``LoadMeta``; by default, this will
    apply recursively to any nested dataclasses. Here's a sample usage of this
    below::

        >>> from dataclass_wizard import LoadMeta
        >>> LoadMeta(key_transform='CAMEL').bind_to(MyClass)
        >>> fromdict(MyClass, {"myStr": "value"})

    """
    try:
        return cls.__dataclass_wizard_from_dict__(d)

    except (AttributeError, TypeError):
        fn = load_func_for_dataclass(cls)
        cls.__dataclass_wizard_from_dict__ = fn  # explicit cache
        return fn(d)


def fromlist(cls: type[T], list_of_dict: list[JSONObject]) -> list[T]:
    """
    Converts a Python list object to a list of dataclass instances.

    Iterates over each dataclass field recursively; lists, dicts, and nested
    dataclasses will likewise be initialized as expected.

    """
    try:
        load = cls.__dataclass_wizard_from_dict__
    except AttributeError:
        load = UNSET

    if load is UNSET:
        load = load_func_for_dataclass(cls)
        cls.__dataclass_wizard_from_dict__ = load  # explicit cache

    return [load(d) for d in list_of_dict]



import warnings
from abc import ABC, abstractmethod
from dataclasses import Field
from typing import (Any, ClassVar, Iterable, Callable, Collection, Sequence)


# added as we can't import from `type_def`, as we run into a circular import.
JSONObject = dict[str, Any]


def type_name(obj: type) -> str:
    """Return the type or class name of an object"""


def show_deprecation_warning(
    fn: Callable | str,
    reason: str,
    fmt: str = "Deprecated function {name} ({reason})."
) -> None:
    """
    Display a deprecation warning for a given function.

    @param fn: Function which is deprecated.
    @param reason: Reason for the deprecation.
    @param fmt: Format string for the name/reason.
    """


class JSONWizardError(ABC, Exception):
    """
    Base error class, for errors raised by this library.
    """

    _TEMPLATE: ClassVar[str]

    _parent_cls: type
    _class_name: str | None
    _default_class_name: str | None

    def class_name(self) -> str | None: ...
    # noinspection PyRedeclaration
    def class_name(self) -> None: ...  # type: ignore[no-redef]

    def parent_cls(self) -> type | None: ...
    # noinspection PyRedeclaration
    def parent_cls(self, value: type | None) -> None: ...  # type: ignore[no-redef]

    @staticmethod
    def name(obj) -> str: ...

    @property
    @abstractmethod
    def message(self) -> str:
        """
        Format and return an error message.
        """

    def __str__(self) -> str: ...


class ParseError(JSONWizardError):
    """
    Base error when an error occurs during the JSON load process.
    """

    _TEMPLATE: str

    obj: Any
    obj_type: type
    phase: str
    ann_type: type | Iterable | None
    base_error: Exception
    kwargs: dict[str, Any]
    _class_name: str | None
    _default_class_name: str | None
    _field_name: str | None
    _json_object: Any | None
    fields: Collection[Field] | None

    def __init__(self, base_err: Exception,
                 obj: Any,
                 ann_type: type | Iterable | None,
                 phase: str,
                 _default_class: type | None = None,
                 _field_name: str | None = None,
                 _json_object: Any = None,
                 **kwargs):
        ...

    @property
    def field_name(self) -> str | None:
        ...

    @property
    def json_object(self):
        ...

    @property
    def message(self) -> str: ...


class ExtraData(JSONWizardError):
    """
    Error raised when extra keyword arguments are passed in to the constructor
    or `__init__()` method of an `EnvWizard` subclass.

    Note that this error class is raised by default, unless a value for the
    `extra` field is specified in the :class:`Meta` class.
    """

    _TEMPLATE: str

    class_name: str
    extra_kwargs: Collection[str]
    field_names: Collection[str]

    def __init__(self,
                 cls: type,
                 extra_kwargs: Collection[str],
                 field_names: Collection[str]):
        ...

    @property
    def message(self) -> str: ...


class MissingFields(JSONWizardError):
    """
    Error raised when unable to create a class instance (most likely due to
    missing arguments)
    """

    _TEMPLATE: str

    obj: JSONObject
    fields: list[str]
    all_fields: tuple[Field, ...]
    missing_fields: Collection[str]
    base_error: Exception | None
    missing_keys: Collection[str] | None
    kwargs: dict[str, Any]
    class_name: str
    parent_cls: type

    def __init__(self, base_err: Exception | None,
                 obj: JSONObject,
                 cls: type,
                 cls_fields: tuple[Field, ...],
                 cls_kwargs: JSONObject | None = None,
                 missing_fields: Collection[str] | None = None,
                 missing_keys: Collection[str] | None = None,
                 **kwargs):
        ...

    @property
    def message(self) -> str: ...


class UnknownKeysError(JSONWizardError):
    """
    Error raised when unknown JSON key(s) are
    encountered in the JSON load process.

    Note that this error class is only raised when the
    `raise_on_unknown_json_key` flag is enabled in
    the :class:`Meta` class.
    """

    _TEMPLATE: str

    unknown_keys: list[str] | str
    obj: JSONObject
    fields: list[str]
    kwargs: dict[str, Any]
    class_name: str

    def __init__(self,
                 unknown_keys: list[str] | str,
                 obj: JSONObject,
                 cls: type,
                 cls_fields: tuple[Field, ...],
                 **kwargs):
        ...

    @property
    @warnings.deprecated('use `unknown_keys` instead')
    def json_key(self) -> list[str] | str: ...

    @property
    def message(self) -> str: ...


# Alias for backwards-compatibility.
UnknownJSONKey = UnknownKeysError


class MissingData(ParseError):
    """
    Error raised when unable to create a class instance, as the JSON object
    is None.
    """

    _TEMPLATE: str

    nested_class_name: str

    def __init__(self, nested_cls: type, **kwargs):
        ...

    @property
    def message(self) -> str: ...


class RecursiveClassError(JSONWizardError):
    """
    Error raised when we encounter a `RecursionError` due to cyclic
    or self-referential dataclasses.
    """

    _TEMPLATE: str

    class_name: str

    def __init__(self, cls: type): ...

    @property
    def message(self) -> str: ...


class InvalidConditionError(JSONWizardError):
    """
    Error raised when a condition is not wrapped in ``SkipIf``.
    """

    _TEMPLATE: str

    class_name: str
    field_name: str

    def __init__(self, cls: type, field_name: str):
        ...

    @property
    def message(self) -> str: ...


class MissingVars(JSONWizardError):
    """
    Error raised when unable to create an instance of a EnvWizard subclass
    (most likely due to missing environment variables in the Environment)

    """
    _TEMPLATE: str

    class_name: str
    fields: str
    def_resolution: str
    init_resolution: str
    prefix: str

    def __init__(self,
                 cls: type,
                 missing_vars: Sequence[tuple[str, str | None, str, Any]]):
        ...

    @property
    def message(self) -> str: ...



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import tzinfo
from typing import (Callable, Type, Dict, Optional, ClassVar, Union,
                    TypeVar, Mapping, Sequence, TYPE_CHECKING, Any, Literal)

from .constants import TAG
from .decorators import cached_class_property
from .enums import DateTimeTo, LetterCase, LetterCasePriority
from .models import Condition

if TYPE_CHECKING:
    from .bases_meta import ALLOWED_MODES, V1HookFn, V1PreDecoder
    from .type_def import FrozenKeys

    V1TypeToHook = Mapping[type, Union[tuple[ALLOWED_MODES, V1HookFn], V1HookFn, None]]

# Create a generic variable that can be 'AbstractMeta', or any subclass.
# Full word as `M` is already defined in another module
META_ = TypeVar('META_', bound='AbstractMeta')
# Use `type` here explicitly, because we will never have an `META_` object.
META = type[META_]


# Create a generic variable that can be 'AbstractMeta', or any subclass.
# Full word as `M` is already defined in another module
ENV_META_ = TypeVar('ENV_META_', bound='AbstractEnvMeta')
# Use `type` here explicitly, because we will never have an `META_` object.
ENV_META = type[ENV_META_]


class ABCOrAndMeta(ABCMeta):
    """
    Metaclass to add class-level :meth:`__or__` and :meth:`__and__` methods
    to a base class of type :type:`M`.

    Ref:
      - https://stackoverflow.com/q/15008807/10237506
      - https://stackoverflow.com/a/57351066/10237506
    """

    def __or__(cls: META, other: META) -> META:
        """
        Merge two Meta configs. Priority will be given to the source config
        present in `cls`, e.g. the first operand in the '|' expression.

        Use case: Merge the Meta configs for two separate dataclasses into a
        single, unified Meta config.
        """
        src = cls
        src_dict = src.__dict__
        other_dict = other.__dict__

        base_dict = {'__slots__': ()}

        # Set meta attributes here.
        if src is AbstractMeta or src is AbstractEnvMeta:
            # Here we can't use `src` because the `bind_to` method isn't
            # defined on the abstract class. Use `other` instead, which
            # *will* be a concrete subclass of `AbstractMeta`.
            src = other
            # noinspection PyTypeChecker
            for k in src.fields_to_merge:
                if k in other_dict:
                    base_dict[k] = other_dict[k]
        else:
            # noinspection PyTypeChecker
            for k in src.fields_to_merge:
                if k in src_dict:
                    base_dict[k] = src_dict[k]
                elif k in other_dict:
                    base_dict[k] = other_dict[k]

        # This mapping won't be updated. Use the src by default.
        for k in src.__special_attrs__:
            if k in src_dict:
                base_dict[k] = src_dict[k]

        new_cls_name = src.__name__
        # Check if the type of the class we want to create is
        # `JSONWizard.Meta` or a subclass. If so, we want to avoid the
        # mandatory `__init_subclass__` call that gets invoked when creating
        # a new class, so use the superclass type instead.
        if src.__is_inner_meta__:
            # In a reversed MRO, the inheritance tree looks like this:
            #   |___ object -> AbstractMeta -> BaseJSONWizardMeta -> ...
            # So here, we want to choose the third-to-last class in the list.
            # noinspection PyUnresolvedReferences
            src = src.__mro__[-3]

        # noinspection PyTypeChecker
        return type(new_cls_name, (src, ), base_dict)

    def __and__(cls: META, other: META) -> META:
        """
        Merge the `other` Meta config into the first one, i.e. `cls`. This
        operation does not create a new class, but instead it modifies the
        source config `cls` in-place; the source will be the first operand in
        the '&' expression.

        Use case: Merge a separate Meta config (for a single dataclass) with
        the first config.
        """
        other_dict = other.__dict__

        # Set meta attributes here.
        # noinspection PyTypeChecker
        for k in cls.all_fields:
            if k in other_dict:
                setattr(cls, k, other_dict[k])

        return cls


class AbstractMeta(metaclass=ABCOrAndMeta):
    """
    Base class definition for the `JSONWizard.Meta` inner class.
    """
    __slots__ = ()

    # A list of class attributes that are exclusive to the Meta config.
    # When merging two Meta configs for a class, these are the only
    # attributes which will *not* be merged.
    __special_attrs__ = frozenset({
        'recursive',
        'json_key_to_field',
        'tag',
    })

    # Class attribute which enables us to detect a `JSONWizard.Meta` subclass.
    __is_inner_meta__ = False

    # Enable Debug mode for more verbose log output.
    #
    # This setting can be a `bool`, `int`, or `str`:
    # - `True` enables debug mode with default verbosity.
    # - A `str` or `int` specifies the minimum log level (e.g., 'DEBUG', 10).
    #
    # Debug mode provides additional helpful log messages, including:
    # - Logging unknown JSON keys encountered during `from_dict` or `from_json`.
    # - Detailed error messages for invalid types during unmarshalling.
    #
    # Note: Enabling Debug mode may have a minor performance impact.
    #
    # @deprecated and will be removed in V1 - Use `v1_debug` instead.
    debug_enabled: ClassVar['bool | int | str'] = False

    # When enabled, a specified Meta config for the main dataclass (i.e. the
    # class on which `from_dict` and `to_dict` is called) will cascade down
    # and be merged with the Meta config for each *nested* dataclass; note
    # that during a merge, priority is given to the Meta config specified on
    # each class.
    #
    # The default behavior is True, so the Meta config (if provided) will
    # apply in a recursive manner.
    recursive: ClassVar[bool] = True

    # True to support cyclic or self-referential dataclasses. For example,
    # the type of a dataclass field in class `A` refers to `A` itself.
    #
    # See https://github.com/rnag/dataclass-wizard/issues/62 for more details.
    recursive_classes: ClassVar[bool] = False

    # True to raise an class:`UnknownJSONKey` when an unmapped JSON key is
    # encountered when `from_dict` or `from_json` is called; an unknown key is
    # one that does not have a known mapping to a dataclass field.
    #
    # The default is to only log a "warning" for such cases, which is visible
    # when `debug_enabled` is true and logging is properly configured.
    raise_on_unknown_json_key: ClassVar[bool] = False

    # A customized mapping of JSON keys to dataclass fields, that is used
    # whenever `from_dict` or `from_json` is called.
    #
    # Note: this is in addition to the implicit field transformations, like
    #   "myStr" -> "my_str"
    #
    # If the reverse mapping is also desired (i.e. dataclass field to JSON
    # key), then specify the "__all__" key as a truthy value. If multiple JSON
    # keys are specified for a dataclass field, only the first one provided is
    # used in this case.
    json_key_to_field: ClassVar[Dict[str, str]] = None

    # How should :class:`time` and :class:`datetime` objects be serialized
    # when converted to a Python dictionary object or a JSON string.
    marshal_date_time_as: ClassVar[Union[DateTimeTo, str]] = None

    # How JSON keys should be transformed to dataclass fields.
    #
    # Note that this only applies to keys which are to be set on dataclass
    # fields; other fields such as the ones for `TypedDict` or `NamedTuple`
    # sub-classes won't be similarly transformed.
    key_transform_with_load: ClassVar[Union[LetterCase, str]] = None

    # How dataclass fields should be transformed to JSON keys.
    #
    # Note that this only applies to dataclass fields; other fields such as
    # the ones for `TypedDict` or `NamedTuple` sub-classes won't be similarly
    # transformed.
    key_transform_with_dump: ClassVar[Union[LetterCase, str]] = None

    # The field name that identifies the tag for a class.
    #
    # When set to a value, an :attr:`TAG` field will be populated in the
    # dictionary object in the dump (serialization) process. When loading
    # (or de-serializing) a dictionary object, the :attr:`TAG` field will be
    # used to load the corresponding dataclass, assuming the dataclass field
    # is properly annotated as a Union type, ex.:
    #   my_data: Union[Data1, Data2, Data3]
    tag: ClassVar[str] = None

    # The dictionary key that identifies the tag field for a class. This is
    # only set when the `tag` field or the `auto_assign_tags` flag is enabled
    # in the `Meta` config for a dataclass.
    #
    # Defaults to '__tag__' if not specified.
    tag_key: ClassVar[str] = TAG

    # Auto-assign the class name as a dictionary "tag" key, for any dataclass
    # fields which are in a `Union` declaration, ex.:
    #   my_data: Union[Data1, Data2, Data3]
    auto_assign_tags: ClassVar[bool] = False

    # Determines whether we should we skip / omit fields with default values
    # (based on the `default` or `default_factory` argument specified for
    # the :func:`dataclasses.field`) in the serialization process.
    skip_defaults: ClassVar[bool] = False

    # Determines the :class:`Condition` to skip / omit dataclass
    # fields in the serialization process.
    skip_if: ClassVar[Condition] = None

    # Determines the condition to skip / omit fields with default values
    # (based on the `default` or `default_factory` argument specified for
    # the :func:`dataclasses.field`) in the serialization process.
    skip_defaults_if: ClassVar[Condition] = None

    # noinspection PyMethodParameters
    @cached_class_property
    def all_fields(cls) -> FrozenKeys:
        """Return a list of all class attributes"""
        return frozenset(AbstractMeta.__annotations__)

    # noinspection PyMethodParameters
    @cached_class_property
    def fields_to_merge(cls) -> FrozenKeys:
        """Return a list of class attributes, minus `__special_attrs__`"""
        return cls.all_fields - cls.__special_attrs__

    @classmethod
    @abstractmethod
    def bind_to(cls, dataclass: Type, create=True, is_default=True):
        """
        Initialize hook which applies the Meta config to `dataclass`, which is
        typically a subclass of :class:`JSONWizard`.

        :param dataclass: A class which has been decorated by the `@dataclass`
          decorator; typically this is a sub-class of :class:`JSONWizard`.
        :param create: When true, a separate loader/dumper will be created
          for the class. If disabled, this will access the root loader/dumper,
          so modifying this should affect global settings across all
          dataclasses that use the JSON load/dump process.
        :param is_default: When enabled, the Meta will be cached as the
          default Meta config for the dataclass. Defaults to true.

        """


class AbstractEnvMeta(metaclass=ABCOrAndMeta):
    """
    Base class definition for the `EnvWizard.Meta` inner class.
    """
    __slots__ = ()

    # A list of class attributes that are exclusive to the Meta config.
    # When merging two Meta configs for a class, these are the only
    # attributes which will *not* be merged.
    __special_attrs__ = frozenset({
        'recursive',
        'debug_enabled',
        'env_var_to_field',
        'tag',
    })

    # Class attribute which enables us to detect a `EnvWizard.Meta` subclass.
    __is_inner_meta__ = False

    # True to enable Debug mode for additional (more verbose) log output.
    #
    # For example, a message is logged with the environment variable that is
    # mapped to each attribute.
    #
    # This also results in more helpful messages during error handling, which
    # can be useful when debugging the cause when values are an invalid type
    # (i.e. they don't match the annotation for the field) when unmarshalling
    # a environ variable values to attributes in an EnvWizard subclass.
    #
    # Note there is a minor performance impact when DEBUG mode is enabled.
    debug_enabled: ClassVar[bool] = False

    # When enabled, a specified Meta config for the main dataclass (i.e. the
    # class on which `from_dict` and `to_dict` is called) will cascade down
    # and be merged with the Meta config for each *nested* dataclass; note
    # that during a merge, priority is given to the Meta config specified on
    # each class.
    #
    # The default behavior is True, so the Meta config (if provided) will
    # apply in a recursive manner.
    recursive: ClassVar[bool] = True

    # `True` to load environment variables from an `.env` file, or a
    # list/tuple of dotenv files.
    #
    # This can also be set to a path to a custom dotenv file, for example:
    #   `path/to/.env.prod`
    #
    # Simply passing in a filename such as `.env.prod` will search the current
    # directory, as well as any parent folders (working backwards to the root
    # directory), until it locates the given file.
    #
    # If multiple files are passed in, later files in the list/tuple will take
    # priority over earlier files.
    #
    # For example, in below the '.env.last' file takes priority over '.env':
    #   env_file = '.env', '.env.last'
    env_file: ClassVar[EnvFilePaths] = None

    # Prefix for all environment variables. Defaults to `None`.
    env_prefix: ClassVar[str] = None

    # secrets_dir: The secret files directory or a sequence of directories. Defaults to `None`.
    secrets_dir: ClassVar[SecretsDirs] = None

    # -- BEGIN Deprecated Fields --

    # The nested env values delimiter. Defaults to `None`.
    # env_nested_delimiter: ClassVar[str] = None

    # A customized mapping of field in the `EnvWizard` subclass to its
    # corresponding environment variable to search for.
    #
    # Note: this is in addition to the implicit field transformations, like
    #   "myStr" -> "my_str"
    field_to_env_var: ClassVar[Dict[str, str]] = None

    # The letter casing priority to use when looking up Env Var Names.
    #
    # The default is `SCREAMING_SNAKE_CASE`.
    key_lookup_with_load: ClassVar[Union[LetterCasePriority, str]] = LetterCasePriority.SCREAMING_SNAKE

    # How `EnvWizard` fields (variables) should be transformed to JSON keys.
    #
    # The default is 'snake_case'.
    key_transform_with_dump: ClassVar[Union[LetterCase, str]] = LetterCase.SNAKE

    # -- END Deprecated Fields --

    # Determines whether we should we skip / omit fields with default values
    # in the serialization process.
    skip_defaults: ClassVar[bool] = False

    # Determines the :class:`Condition` to skip / omit dataclass
    # fields in the serialization process.
    skip_if: ClassVar[Condition] = None

    # Determines the condition to skip / omit fields with default values
    # (based on the `default` or `default_factory` argument specified for
    # the :func:`dataclasses.field`) in the serialization process.
    skip_defaults_if: ClassVar[Condition] = None

    # The field name that identifies the tag for a class.
    #
    # When set to a value, an :attr:`TAG` field will be populated in the
    # dictionary object in the dump (serialization) process. When loading
    # (or de-serializing) a dictionary object, the :attr:`TAG` field will be
    # used to load the corresponding dataclass, assuming the dataclass field
    # is properly annotated as a Union type, ex.:
    #   my_data: Union[Data1, Data2, Data3]
    tag: ClassVar[str] = None

    # The dictionary key that identifies the tag field for a class. This is
    # only set when the `tag` field or the `auto_assign_tags` flag is enabled
    # in the `Meta` config for a dataclass.
    #
    # Defaults to '__tag__' if not specified.
    tag_key: ClassVar[str] = TAG

    # Auto-assign the class name as a dictionary "tag" key, for any dataclass
    # fields which are in a `Union` declaration, ex.:
    #   my_data: Union[Data1, Data2, Data3]
    auto_assign_tags: ClassVar[bool] = False

    # noinspection PyMethodParameters
    @cached_class_property
    def all_fields(cls) -> FrozenKeys:
        """Return a list of all class attributes"""
        return frozenset(AbstractEnvMeta.__annotations__)

    # noinspection PyMethodParameters
    @cached_class_property
    def fields_to_merge(cls) -> FrozenKeys:
        """Return a list of class attributes, minus `__special_attrs__`"""
        return cls.all_fields - cls.__special_attrs__

    @classmethod
    @abstractmethod
    def bind_to(cls, env_class: Type, create=True, is_default=True):
        """
        Initialize hook which applies the Meta config to `env_class`, which is
        typically a subclass of :class:`EnvWizard`.

        :param env_class: A sub-class of :class:`EnvWizard`.
        :param create: When true, a separate loader/dumper will be created
          for the class. If disabled, this will access the root loader/dumper,
          so modifying this should affect global settings across all
          dataclasses that use the JSON load/dump process.
        :param is_default: When enabled, the Meta will be cached as the
          default Meta config for the dataclass. Defaults to true.

        """


class BaseLoadHook:
    """
    Container class for type hooks.
    """
    __slots__ = ()

    __LOAD_HOOKS__: ClassVar[Dict[Type, Callable]] = None

    def __init_subclass__(cls):
        super().__init_subclass__()
        # (Re)assign the dict object so we have a fresh copy per class
        cls.__LOAD_HOOKS__ = {}

    @classmethod
    def register_load_hook(cls, typ: Type, func: Callable):
        """Registers the hook for a type, on the default loader by default."""
        cls.__LOAD_HOOKS__[typ] = func

    @classmethod
    def get_load_hook(cls, typ: Type) -> Optional[Callable]:
        """Retrieves the hook for a type, if one exists."""
        return cls.__LOAD_HOOKS__.get(typ)


class BaseDumpHook:
    """
    Container class for type hooks.
    """
    __slots__ = ()

    __DUMP_HOOKS__: ClassVar[Dict[Type, Callable]] = None

    def __init_subclass__(cls):
        super().__init_subclass__()
        # (Re)assign the dict object so we have a fresh copy per class
        cls.__DUMP_HOOKS__ = {}

    @classmethod
    def register_dump_hook(cls, typ: Type, func: Callable):
        """Registers the hook for a type, on the default dumper by default."""
        cls.__DUMP_HOOKS__[typ] = func

    @classmethod
    def get_dump_hook(cls, typ: Type) -> Optional[Callable]:
        """Retrieves the hook for a type, if one exists."""
        return cls.__DUMP_HOOKS__.get(typ)



from collections.abc import Callable
from datetime import date, datetime, time, timedelta, timezone, tzinfo
from typing import Any, AnyStr
from typing import Callable as _Callable

from _typeshed import Incomplete

from ._type_def import E, N

__all__ = ['TRUTHY_VALUES', 'as_int', 'as_datetime', 'as_date', 'as_time', 'as_timedelta', 'datetime_to_timestamp', 'as_collection', 'as_list', 'as_dict', 'as_enum']

TRUTHY_VALUES: frozenset
def as_int(o: float | bool, tp: type, base_type: type[int] = ...): ...
def as_datetime(o: int | float | datetime, _from_timestamp: Callable[[float, tzinfo], datetime], _tz: Incomplete | None = ...): ...
def as_date(o: int | float | date, _from_timestamp: Callable[[float, tzinfo], datetime], _tz: Incomplete | None = ..., _cls: type[date] = ...): ...
def as_time(o: time | Any, base_type: type[time]): ...
# noinspection PyTypeHints
def as_timedelta(o: str | N | timedelta, base_type: type[timedelta] = ..., default: Incomplete | None = ..., raise_: bool = ...): ...
def datetime_to_timestamp(dt: datetime, assume_naive_tz: timezone) -> int: ...
def as_collection(v: Any, *, strip: bool = ...) -> Any: ...
def as_list(v: Any, *, sep: str = ..., strip: bool = ..., drop_empty: bool = ..., json_enabled: bool = ...) -> Any: ...
def as_dict(v: Any, *, sep: str = ..., kv_sep: str = ..., strip: bool = ..., drop_empty: bool = ..., json_enabled: bool = ..., allow_bare_keys: bool = ...) -> Any: ...
def as_enum(o: AnyStr | N, base_type: type[E], lookup_func: _Callable = ..., transform_func: _Callable = ..., raise_: bool = ...) -> E | None: ...  # type: ignore[valid-type]



import os
from dataclasses import MISSING

import pytest

from dataclass_wizard.v0.environ.lookups import *


@pytest.mark.parametrize(
    'string,expected',
    [
        ('device_type', 'devicetype'),
        ('isACamelCasedWORD', 'isacamelcasedword'),
        ('ATitledWordToTESTWith', 'atitledwordtotestwith'),
        ('not-a-tester', 'notatester'),
        ('helloworld', 'helloworld'),
        ('A', 'a'),
        ('TESTing_if_thisWorks', 'testingifthisworks'),
        ('a_B_Cde_fG_hi', 'abcdefghi'),
        ('How_-Are-_YoUDoing__TeST', 'howareyoudoingtest'),
    ]
)
def test_clean(string, expected):
    assert clean(string) == expected


def test_lookup_exact():
    assert lookup_exact('abc-this-key-shouldnt-exist') is MISSING
    assert lookup_exact(('abc-this-key-shouldnt-exist', )) is MISSING


def test_reload_when_not_accessed_cleaned_to_env():
    # save current value
    current_val = Env._accessed_cleaned_to_env

    Env._accessed_cleaned_to_env = False
    Env.reload()

    # don't forget to reset it
    Env._accessed_cleaned_to_env = current_val


def test_with_snake_case():
    var = 'my_test_string_1'
    assert with_snake_case(var) is MISSING

    os.environ['MY_TEST_STRING_1'] = 'hello world'
    Env.reload()
    assert with_snake_case(var) == 'hello world'

    os.environ[var] = 'testing 123'
    Env.reload()
    assert with_snake_case(var) == 'testing 123'


def test_with_pascal_or_camel_case():
    var = 'MyTestString2'
    assert with_pascal_or_camel_case(var) is MISSING

    os.environ['my_test_string2'] = 'testing 123'
    Env.reload()
    assert with_pascal_or_camel_case(var) == 'testing 123'

    os.environ['MY_TEST_STRING2'] = 'hello world'
    Env.reload()
    assert with_pascal_or_camel_case(var) == 'hello world'

    if os.name == 'nt':
        # Windows: var names are automatically converted
        # to upper case when saved to `os.environ`
        return

    os.environ[var] = 'hello world !!'
    Env.reload()
    assert with_pascal_or_camel_case(var) == 'hello world !!'



from functools import wraps
from typing import Any, Dict, Type, Callable, Union, TypeVar, cast

from .constants import SINGLE_ARG_ALIAS, IDENTITY
from .errors import ParseError


T = TypeVar('T')


# noinspection PyPep8Naming
class cached_class_property(object):
    """
    Descriptor decorator implementing a class-level, read-only property,
    which caches the attribute on-demand on the first use.

    Credits: https://stackoverflow.com/a/4037979/10237506
    """
    def __init__(self, func):
        self.__func__ = func
        self.__attr_name__ = func.__name__

    def __get__(self, instance, cls=None):
        """This method is only called the first time, to cache the value."""
        if cls is None:
            cls = type(instance)

        # Build the attribute.
        attr = self.__func__(cls)

        # Cache the value; hide ourselves.
        setattr(cls, self.__attr_name__, attr)

        return attr


class cached_property(object):
    """
    Descriptor decorator implementing an instance-level, read-only property,
    which caches the attribute on-demand on the first use.
    """
    def __init__(self, func):
        self.__func__ = func
        self.__attr_name__ = func.__name__

    def __get__(self, instance, cls=None):
        """This method is only called the first time, to cache the value."""
        # Build the attribute.
        attr = self.__func__(instance)

        # Cache the value; hide ourselves.
        setattr(instance, self.__attr_name__, attr)

        return attr


def try_with_load(load_fn: Callable):
    """Try to call a load hook, catch and re-raise errors as a ParseError.

    Note: this function will be recursively called on all load hooks for a
    dataclass, when `debug_mode` is enabled for the dataclass.

    :param load_fn: The load hook, can be a regular callable, a single-arg
      alias, or an identity function.
    :return: The decorated load hook.
    """
    try:  # Check if it's a single-argument function, ex. float(...)
        single_arg_alias_func = getattr(load_fn, SINGLE_ARG_ALIAS)

    except AttributeError:
        # Check if it's an identity function, ex. lambda o: o
        if hasattr(load_fn, IDENTITY):
            # These are basically do-nothing callables, so we don't need to
            # decorate them.
            return load_fn

        @wraps(load_fn)
        def new_func(o: Any, base_type: Type, *args, **kwargs):
            try:
                return load_fn(o, base_type, *args, **kwargs)

            except ParseError as e:
                # This means that a nested load hook raised an exception.
                # Therefore, to help with debugging we should print the name
                # of the outer load hook and the original object.
                e.kwargs['load_hook'] = load_fn.__name__
                e.obj = o
                # Re-raise the original error
                raise

            except Exception as e:
                raise ParseError(e, o, base_type, 'load', load_hook=load_fn.__name__)

        return new_func

    else:
        # fix: avoid re-decoration when DEBUG mode is enabled multiple
        # times (i.e. on more than one class)
        if hasattr(load_fn, '__decorated__'):
            return load_fn

        # If it's a string value, we don't know the name of the load hook
        # function (method) beforehand.
        if isinstance(single_arg_alias_func, str):
            alias = single_arg_alias_func
            f_locals = {}
        else:
            alias = single_arg_alias_func.__name__
            f_locals = {alias: single_arg_alias_func}

        wrapped_fn = f'{try_with_load_with_single_arg.__name__}' \
                     f'(original_fn, {alias}, base_type)'

        setattr(load_fn, '__decorated__', True)
        setattr(load_fn, SINGLE_ARG_ALIAS, wrapped_fn)
        setattr(load_fn, 'f_locals', f_locals)

        return load_fn


def try_with_load_with_single_arg(original_fn: Callable,
                                  single_arg_load_fn: Callable,
                                  base_type: Type):
    """Similar to :func:`try_with_load`, but for single-arg alias functions.

    :param original_fn: The original load hook (function)
    :param single_arg_load_fn: The single-argument load hook
    :param base_type: The annotated (or desired) type
    :return: The decorated load hook.
    """
    @wraps(single_arg_load_fn)
    def new_func(o: Any):
        try:
            return single_arg_load_fn(o)

        except ParseError as e:
            # This means that a nested load hook raised an exception.
            # Therefore, to help with debugging we should print the name
            # of the outer load hook and the original object.
            e.kwargs['load_hook'] = original_fn.__name__
            e.obj = o
            # Re-raise the original error
            raise

        except Exception as e:
            raise ParseError(e, o, base_type, 'load', load_hook=original_fn.__name__)

    return new_func


def _alias(default: Callable) -> Callable[[T], T]:
    """
    Decorator which re-assigns a function `_f` to point to `default` instead.
    Since global function calls in Python are somewhat expensive, this is
    mainly done to reduce a bit of overhead involved in the functions calls.

    For example, consider the below example::

        def f2(o):
            return o

        def f1(o):
            return f2(o)

    Calling function `f1` will incur some additional overhead, as opposed to
    simply calling `f2`.

    Now assume we wrap `f1` with the `_alias` decorator::

        def f2(o):
            return o

        @_alias(f2)
        def f1(o):
            ...

    This will essentially perform the assignment of `f1 = f2`, so calling
    `f1()` in this case has no additional function overhead, as opposed to
    just calling `f2()`.
    """

    def new_func(_f: T) -> T:
        return cast(T, default)

    return new_func


def _single_arg_alias(alias_func: Union[Callable, str] = None):
    """
    Decorator which wraps a function to set the :attr:`SINGLE_ARG_ALIAS` on
    a function `f`, which is an alias function that takes only one argument.
    This is useful mainly so that other functions can access this attribute,
    and can opt to call it instead of function `f`.
    """

    def new_func(f):
        setattr(f, SINGLE_ARG_ALIAS, alias_func)
        return f

    return new_func


def _identity(_f: Callable = None, id: Union[object, str] = None):
    """
    Decorator which wraps a function to set the :attr:`IDENTITY` on a function
    `f`, indicating that this is an identity function that returns its first
    argument. This is useful mainly so that other functions can access this
    attribute, and can opt to call it instead of function `f`.
    """

    def new_func(f):
        setattr(f, IDENTITY, id)
        return f

    return new_func(_f) if _f else new_func


def resolve_alias_func(f: Callable,
                       _locals: Dict = None,
                       raise_=False) -> Callable:
    """
    Resolve the underlying single-arg alias function for `f`, using the
    provided function locals (which will be a dict). If `f` does not have an
    associated alias function, we return `f` itself.

    :raises AttributeError: If `raise_` is true and `f` is not a single-arg
      alias function.
    """

    try:
        single_arg_alias_func = getattr(f, SINGLE_ARG_ALIAS)

    except AttributeError:
        if raise_:
            raise
        return f

    else:
        if isinstance(single_arg_alias_func, str) and _locals is not None:
            try:
                return _locals[single_arg_alias_func]
            except KeyError:
                # This is only the case when debug mode is enabled, so the
                # string will be like 'try_with_load_with_single_arg(...)'
                _locals['original_fn'] = f
                f_locals = getattr(f, 'f_locals', None)
                if f_locals:
                    _locals.update(f_locals)

                return eval(single_arg_alias_func, globals(), _locals)

        return single_arg_alias_func



__all__ = [
    'Buffer',
    'Unpack',
    'PyForwardRef',
    'PyProtocol',
    'PyDeque',
    'PyTypedDict',
    'PyRequired',
    'PyNotRequired',
    'PyReadOnly',
    'PyLiteralString',
    'FrozenKeys',
    'DefFactory',
    'NoneType',
    'ExplicitNullType',
    'ExplicitNull',
    'JSONList',
    'JSONObject',
    'ListOfJSONObject',
    'JSONValue',
    'FileType',
    'EnvFileType',
    'StrCollection',
    'ParseFloat',
    'Encoder',
    'FileEncoder',
    'Decoder',
    'FileDecoder',
    'NUMBERS',
    'T',
    'E',
    'U',
    'M',
    'NT',
    'DT',
    'DD',
    'N',
    'S',
    'LT',
    'LSQ',
    'FREF',
    'dataclass_transform',
]

from collections import deque, defaultdict
from datetime import date, time, datetime
from enum import Enum
from os import PathLike
from typing import (
    Any, TypeVar, Sequence, Mapping,
    Union, NamedTuple, Callable, AnyStr, TextIO, BinaryIO,
    Deque as PyDeque,
    ForwardRef as PyForwardRef,
    Protocol as PyProtocol,
    TypedDict as PyTypedDict, Iterable, Collection,
)
from uuid import UUID

from .constants import PY310_OR_ABOVE, PY311_OR_ABOVE, PY313_OR_ABOVE, PY312_OR_ABOVE

# The class of the `None` singleton, cached for re-usability
if PY310_OR_ABOVE:
    # https://docs.python.org/3/library/types.html#types.NoneType
    from types import NoneType
else:
    # "Cannot assign to a type"
    NoneType = type(None)  # type: ignore[misc]

# Type check for numeric types - needed because `bool` is technically
# a Number.
NUMBERS = int, float

# Generic type
T = TypeVar('T')
TT = TypeVar('TT')

# Enum subclass type
E = TypeVar('E', bound=Enum)

# UUID subclass type
U = TypeVar('U', bound=UUID)

# Mapping type
M = TypeVar('M', bound=Mapping)

# NamedTuple type
NT = TypeVar('NT', bound=NamedTuple)

# Date, time, or datetime type
DT = TypeVar('DT', date, time, datetime)

# DefaultDict type
DD = TypeVar('DD', bound=defaultdict)

# Numeric type
N = Union[int, float]

# Sequence type
S = TypeVar('S', bound=Sequence)

# List or Tuple type
LT = TypeVar('LT', list, tuple)

# List, Set, or Deque (Double ended queue) type
LSQ = TypeVar('LSQ', list, set, frozenset, deque)

# A fixed set of key names
FrozenKeys = frozenset[str]

# Default factory type, assuming a no-args constructor
DefFactory = Callable[[], T]

# Valid collection types in JSON.
JSONList = list[Any]
JSONObject = dict[str, Any]
ListOfJSONObject = list[JSONObject]

# Valid value types in JSON.
JSONValue = Union[None, str, bool, int, float, JSONList, JSONObject]

# File-type argument, compatible with the type of `file` for `open`
FileType = Union[str, bytes, PathLike, int]

# DotEnv file-type argument (string, tuple of string, boolean, or None)
EnvFileType = Union[bool, FileType, Iterable[FileType], None]

# Type for a string or a collection of strings.
StrCollection = Union[str, Collection[str]]

# Python 3.11 introduced `Required` and `NotRequired` wrappers for
# `TypedDict` fields (PEP 655). Python 3.9+ users can import the
# wrappers from `typing_extensions`.

if PY313_OR_ABOVE:  # pragma: no cover
    from collections.abc import Buffer

    from typing import (Unpack,
                        Required as PyRequired,
                        NotRequired as PyNotRequired,
                        ReadOnly as PyReadOnly,
                        LiteralString as PyLiteralString,
                        dataclass_transform)
elif PY311_OR_ABOVE:  # pragma: no cover
    if PY312_OR_ABOVE:
        from collections.abc import Buffer
    else:
        from typing_extensions import Buffer

    from typing import (Unpack,
                        Required as PyRequired,
                        NotRequired as PyNotRequired,
                        LiteralString as PyLiteralString,
                        dataclass_transform)
    from typing_extensions import ReadOnly as PyReadOnly
else:
    from typing_extensions import (Unpack,
                                   Buffer,
                                   Required as PyRequired,
                                   NotRequired as PyNotRequired,
                                   ReadOnly as PyReadOnly,
                                   LiteralString as PyLiteralString,
                                   dataclass_transform)

# Forward references can be either strings or explicit `ForwardRef` objects.
# noinspection SpellCheckingInspection
FREF = TypeVar('FREF', str, PyForwardRef)


class ExplicitNullType:
    __slots__ = ()  # Saves memory by preventing the creation of instance dictionaries

    # Class-level instance variable for singleton control
    _instance: "ExplicitNullType | None" = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ExplicitNullType, cls).__new__(cls)
        return cls._instance

    def __bool__(self):
        return False

    def __repr__(self):
        return '<ExplicitNull>'


# Create the singleton instance
ExplicitNull = ExplicitNullType()

# Type annotations
ParseFloat = Callable[[str], Any]


class Encoder(PyProtocol):
    """
    Represents an encoder for Python object -> JSON, e.g. analogous to
    `json.dumps`
    """

    def __call__(self, obj: Union[JSONObject, JSONList],
                 /,
                 *args,
                 **kwargs) -> str:
        ...


class FileEncoder(PyProtocol):
    """
    Represents an encoder for Python object -> JSON file, e.g. analogous to
    `json.dump`
    """

    def __call__(self, obj: Union[JSONObject, JSONList],
                 file: Union[TextIO, BinaryIO],
                 **kwargs) -> AnyStr:
        ...


class Decoder(PyProtocol):
    """
    Represents a decoder for JSON -> Python object, e.g. analogous to
    `json.loads`
    """

    def __call__(self, s: AnyStr,
                 **kwargs) -> Union[JSONObject, ListOfJSONObject]:
        ...


class FileDecoder(PyProtocol):
    """
    Represents a decoder for JSON file -> Python object, e.g. analogous to
    `json.load`
    """
    def __call__(self, file: Union[TextIO, BinaryIO],
                 **kwargs) -> Union[JSONObject, ListOfJSONObject]:
        ...



import logging
from dataclasses import dataclass, asdict
from timeit import timeit
from typing import Optional, TypeVar

import dataclass_factory
import pytest
from dataclasses_json import DataClassJsonMixin
from jsons import JsonSerializable
from dacite import from_dict as dacite_from_dict
from pydantic import BaseModel
import marshmallow
import attr
import mashumaro

from dataclass_wizard import JSONWizard, LoadMeta
from dataclass_wizard._type_utils import create_new_class
from dataclass_wizard.constants import PY314_OR_ABOVE
from dataclass_wizard.utils._string_case import to_snake_case

log = logging.getLogger(__name__)

# Dataclass for the test
@dataclass
class MyClass:
    my_str: str
    my_int: int
    my_bool: Optional[bool]

# Add Pydantic Model
class MyClassPydantic(BaseModel):
    my_str: str
    my_int: int
    my_bool: Optional[bool]

# Marshmallow Schema
class MyClassSchema(marshmallow.Schema):
    my_str = marshmallow.fields.Str()
    my_int = marshmallow.fields.Int()
    my_bool = marshmallow.fields.Bool()

# attrs Class
@attr.s
class MyClassAttrs:
    my_str = attr.ib(type=str)
    my_int = attr.ib(type=int)
    my_bool = attr.ib(type=Optional[bool])

# Mashumaro Model
@dataclass
class MyClassMashumaro(mashumaro.DataClassDictMixin):
    my_str: str
    my_int: int
    my_bool: Optional[bool]

# Model for `dataclass-wizard`
WizType = TypeVar("WizType", MyClass, JSONWizard)
# Model for `jsons`
JsonsType = TypeVar("JsonsType", MyClass, JsonSerializable)
# Model for `dataclasses-json`
DJType = TypeVar("DJType", MyClass, DataClassJsonMixin)
# Factory for `dataclass-factory`
factory = dataclass_factory.Factory()

MyClassWizard: WizType = create_new_class(MyClass, (MyClass, JSONWizard), "Wizard")
MyClassDJ: DJType = create_new_class(MyClass, (MyClass, DataClassJsonMixin), "DJ")
MyClassJsons: JsonsType = create_new_class(MyClass, (MyClass, JsonSerializable), "Jsons")

# Enable experimental `v1` mode for optimized de/serialization
LoadMeta(v1=True).bind_to(MyClassWizard)


@pytest.fixture(scope="session")
def data():
    return {
        "my_str": "hello world!",
        "my_int": 21,
        "my_bool": True,
    }

def test_load(data, n):
    """
    [ RESULTS]
    platform darwin -- Python 3.13.11, pytest-8.3.4, pluggy-1.6.0

    benchmarks.simple.simple - [INFO] dataclass-wizard     0.029298
    benchmarks.simple.simple - [INFO] dataclass-factory    0.100123
    benchmarks.simple.simple - [INFO] dataclasses-json     3.530623
    benchmarks.simple.simple - [INFO] jsons                4.920980
    benchmarks.simple.simple - [INFO] dacite               0.612328
    benchmarks.simple.simple - [INFO] pydantic             0.063677
    benchmarks.simple.simple - [INFO] marshmallow          2.197097
    benchmarks.simple.simple - [INFO] attrs                0.020192
    benchmarks.simple.simple - [INFO] mashumaro            0.040619
    """
    g = globals().copy()
    g.update(locals())

    # Add dacite and pydantic benchmarks
    log.info("dataclass-wizard     %f",
             timeit("MyClassWizard.from_dict(data)", globals=g, number=n))
    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        log.info("dataclass-factory    %f",
                 timeit("factory.load(data, MyClass)", globals=g, number=n))
    log.info("dataclasses-json     %f",
             timeit("MyClassDJ.from_dict(data)", globals=g, number=n))
    log.info("jsons                %f",
             timeit("MyClassJsons.load(data)", globals=g, number=n))
    log.info("dacite               %f",
             timeit("dacite_from_dict(MyClass, data)", globals=g, number=n))
    log.info("pydantic             %f",
             timeit("MyClassPydantic(**data)", globals=g, number=n))
    log.info("marshmallow          %f",
             timeit("MyClassSchema().load(data)", globals=g, number=n))
    log.info("attrs                %f",
             timeit("MyClassAttrs(**data)", globals=g, number=n))
    log.info("mashumaro            %f",
             timeit("MyClassMashumaro.from_dict(data)", globals=g, number=n))

    # Assert the dataclass instances have the same values for all fields.
    c1 = MyClassWizard.from_dict(data)
    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        c2 = factory.load(data, MyClass)
    c3 = MyClassDJ.from_dict(data)
    c4 = MyClassJsons.load(data)
    c5 = dacite_from_dict(MyClass, data)
    c6 = MyClassPydantic(**data)
    c7 = MyClassSchema().load(data)
    c8 = MyClassAttrs(**data)
    c9 = MyClassMashumaro.from_dict(data)

    assert c1.__dict__ == c3.__dict__ == c4.__dict__ == c5.__dict__ == c6.model_dump() == c7 == c8.__dict__ == c9.to_dict()

    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        assert c1.__dict__ == c2.__dict__


def test_dump(data, n):
    """
    [ RESULTS]
    platform darwin -- Python 3.13.11, pytest-8.3.4, pluggy-1.6.0

    benchmarks.simple.simple - [INFO] dataclass-wizard     0.010870
    benchmarks.simple.simple - [INFO] asdict (dataclasses) 0.085224
    benchmarks.simple.simple - [INFO] dataclass-factory    0.070084
    benchmarks.simple.simple - [INFO] dataclasses-json     1.272380
    benchmarks.simple.simple - [INFO] jsons                5.980036
    benchmarks.simple.simple - [INFO] dacite (not applicable) -- skipped
    benchmarks.simple.simple - [INFO] pydantic             0.079050
    benchmarks.simple.simple - [INFO] marshmallow          0.000489
    benchmarks.simple.simple - [INFO] attrs                0.054118
    benchmarks.simple.simple - [INFO] mashumaro            0.008982
    """

    c1 = MyClassWizard.from_dict(data)
    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        c2 = factory.load(data, MyClass)
    c3 = MyClassDJ.from_dict(data)
    c4 = MyClassJsons.load(data)
    c5 = dacite_from_dict(MyClass, data)
    c6 = MyClassPydantic(**data)
    c7 = MyClassSchema().load(data)
    c8 = MyClassAttrs(**data)
    c9 = MyClassMashumaro.from_dict(data)

    g = globals().copy()
    g.update(locals())

    log.info("dataclass-wizard     %f",
             timeit("c1.to_dict()", globals=g, number=n))
    log.info("asdict (dataclasses) %f",
             timeit("asdict(c1)", globals=g, number=n))
    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        log.info("dataclass-factory    %f",
                 timeit("factory.dump(c2, MyClass)", globals=g, number=n))
    log.info("dataclasses-json     %f",
             timeit("c3.to_dict()", globals=g, number=n))
    log.info("jsons                %f",
             timeit("c4.dump()", globals=g, number=n))
    log.info("dacite (not applicable) -- skipped")
    log.info("pydantic             %f",
             timeit("c6.model_dump()", globals=g, number=n))
    log.info("marshmallow          %f",
             timeit("c7", globals=g, number=n))
    log.info("attrs                %f",
             timeit("attr.asdict(c8)", globals=g, number=n))
    log.info("mashumaro            %f",
             timeit("c9.to_dict()", globals=g, number=n))

    # Assert the dict objects which are the result of `to_dict` are all equal.
    c1_dict = {to_snake_case(f): fval for f, fval in c1.to_dict().items()}

    assert c1_dict == c3.to_dict() == c4.dump() == c6.model_dump() == attr.asdict(c8) == c9.to_dict()

    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        assert c1_dict == factory.dump(c2, MyClass)



import types
from collections.abc import MutableMapping
from typing import Any

class LazyLoader(types.ModuleType):
    def __init__(
        self,
        parent_module_globals: MutableMapping[str, Any],
        name: str,
        extra: str | None = ...,
        local_name: str | None = ...,
        warning: str | None = ...,
    ) -> None: ...
    def load(self) -> types.ModuleType: ...
    def __getattr__(self, item: str) -> Any: ...
    def __dir__(self) -> list[str]: ...
